#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Solubility laws for sulfur species

For every law there should be a test in the test suite.
"""

import logging

import equinox as eqx
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from atmodeller.solubility.core import Solubility
from atmodeller.utilities import unit_conversion

try:
    from typing import override  # type: ignore valid for Python 3.12+
except ImportError:
    from typing_extensions import override  # Python 3.11 and earlier

logger: logging.Logger = logging.getLogger(__name__)


class _S2_sulfate_andesite_boulliung23(Solubility):
    """Sulfur as sulfate SO4^2-/S^6+ in andesite :cite:p:`BW22,BW23corr`

    Using the first equation in the abstract of :cite:t:`BW22` and the corrected expression for
    sulfate capacity (C_S6+) in :cite:t:`BW23corr`. Composition for andesite from
    :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K for silicate melts
    equilibrated with Air/SO2 mixtures.
    """

    @override
    @eqx.filter_jit
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        del kwargs
        logcs: Array = -12.948 + (31586.2393 / jnp.asarray(temperature))
        logso4_wtp: Array = logcs + (0.5 * jnp.log10(fugacity)) + (1.5 * jnp.log10(fO2))
        so4_wtp: Array = jnp.power(10, logso4_wtp)
        s_wtp: Array = so4_wtp * (32.065 / 96.06)
        ppmw: Array = s_wtp * unit_conversion.percent_to_ppm

        return ppmw


S2_sulfate_andesite_boulliung23: Solubility = _S2_sulfate_andesite_boulliung23()
"""Sulfur as sulfate SO4^2-/S^6+ in andesite :cite:p:`BW22,BW23corr`

Using the first equation in the abstract of :cite:t:`BW22` and the corrected expression for
sulfate capacity (C_S6+) in :cite:t:`BW23corr`. Composition for andesite from
:cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K for silicate melts
equilibrated with Air/SO2 mixtures.
"""


class _S2_sulfide_andesite_boulliung23(Solubility):
    """Sulfur as sulfide (S^2-) in andesite :cite:p:`BW23`

    Using expressions in the abstract for S wt.% and sulfide capacity (C_S2-). Composition
    for andesite from :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K in a
    controlled CO-CO2-SO2 atmosphere fO2 conditions were greater than 1 log unit below FMQ.
    """

    @override
    @eqx.filter_jit
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        del kwargs
        logcs: Array = 0.225 - (8921.0927 / jnp.asarray(temperature))
        logs_wtp: Array = logcs - (0.5 * (jnp.log10(fO2) - jnp.log10(fugacity)))
        s_wtp: Array = jnp.power(10, logs_wtp)
        ppmw: Array = s_wtp * unit_conversion.percent_to_ppm

        return ppmw


S2_sulfide_andesite_boulliung23: Solubility = _S2_sulfide_andesite_boulliung23()
"""Sulfur as sulfide (S^2-) in andesite :cite:p:`BW23`

Using expressions in the abstract for S wt.% and sulfide capacity (C_S2-). Composition
for andesite from :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K in a
controlled CO-CO2-SO2 atmosphere fO2 conditions were greater than 1 log unit below FMQ.
"""


class _S2_andesite_boulliung23(Solubility):
    """S2 in andesite accounting for both sulfide and sulfate :cite:p:`BW22,BW23corr,BW23`"""

    _sulfide: Solubility = eqx.field(init=False)
    _sulfate: Solubility = eqx.field(init=False)

    def __post_init__(self):
        self._sulfide = S2_sulfide_andesite_boulliung23
        self._sulfate = S2_sulfate_andesite_boulliung23

    @override
    @eqx.filter_jit
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        concentration: ArrayLike = self._sulfide.concentration(
            fugacity, temperature=temperature, fO2=fO2, **kwargs
        )
        concentration = concentration + self._sulfate.concentration(
            fugacity, temperature=temperature, fO2=fO2, **kwargs
        )

        return concentration


S2_andesite_boulliung23: Solubility = _S2_andesite_boulliung23()
"""S2 in andesite accounting for both sulfide and sulfate :cite:p:`BW22,BW23corr,BW23`"""


class _S2_sulfate_basalt_boulliung23(Solubility):
    """Sulfur in basalt as sulfate, SO4^2-/S^6+ :cite:p:`BW22,BW23corr`

    Using the first equation in the abstract and the corrected expression for sulfate capacity
    (C_S6+) in :cite:t:`BW23corr`. Composition for Basalt from :cite:t:`BW22{Table 1}`. Experiments
    conducted at 1 atm pressure, temperatures from 1473-1773 K for silicate melts equilibrated with
    Air/SO2 mixtures.
    """

    @override
    @eqx.filter_jit
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        del kwargs
        logcs: Array = -12.948 + (32333.5635 / jnp.asarray(temperature))
        logso4_wtp: Array = logcs + (0.5 * jnp.log10(fugacity)) + (1.5 * jnp.log10(fO2))
        so4_wtp: Array = jnp.power(10, logso4_wtp)
        s_wtp: Array = so4_wtp * (32.065 / 96.06)
        ppmw: Array = s_wtp * unit_conversion.percent_to_ppm

        return ppmw


S2_sulfate_basalt_boulliung23: Solubility = _S2_sulfate_basalt_boulliung23()
"""Sulfur in basalt as sulfate, SO4^2-/S^6+ :cite:p:`BW22,BW23corr`

Using the first equation in the abstract and the corrected expression for sulfate capacity
(C_S6+) in :cite:t:`BW23corr`. Composition for Basalt from :cite:t:`BW22{Table 1}`. Experiments
conducted at 1 atm pressure, temperatures from 1473-1773 K for silicate melts equilibrated with
Air/SO2 mixtures.
"""


class _S2_sulfide_basalt_boulliung23(Solubility):
    """Sulfur in basalt as sulfide (S^2-) :cite:p:`BW23`

    Using expressions in the abstract for S wt% and sulfide capacity (C_S2-). Composition for
    basalt from :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm pressure and temperatures
    from 1473-1773 K in a controlled CO-CO2-SO2 atmosphere fO2 conditions were greater than 1 log
    unit below FMQ.
    """

    @override
    @eqx.filter_jit
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        del kwargs
        logcs: Array = 0.225 - (8045.7465 / jnp.asarray(temperature))
        logs_wtp: Array = logcs - (0.5 * (jnp.log10(fO2) - jnp.log10(fugacity)))
        s_wtp: Array = jnp.power(10, logs_wtp)
        ppmw: Array = s_wtp * unit_conversion.percent_to_ppm

        return ppmw


S2_sulfide_basalt_boulliung23: Solubility = _S2_sulfide_basalt_boulliung23()
"""Sulfur in basalt as sulfide (S^2-) :cite:p:`BW23`

Using expressions in the abstract for S wt% and sulfide capacity (C_S2-). Composition for
basalt from :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm pressure and temperatures
from 1473-1773 K in a controlled CO-CO2-SO2 atmosphere fO2 conditions were greater than 1 log
unit below FMQ.
"""


class _S2_basalt_boulliung23(Solubility):
    """Sulfur in basalt due to sulfide and sulfate dissolution :cite:p:`BW22,BW23corr,BW23`"""

    _sulfide: Solubility = eqx.field(init=False)
    _sulfate: Solubility = eqx.field(init=False)

    def __post_init__(self):
        self._sulfide = S2_sulfide_basalt_boulliung23
        self._sulfate = S2_sulfate_basalt_boulliung23

    @override
    @eqx.filter_jit
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        concentration: ArrayLike = self._sulfide.concentration(
            fugacity, temperature=temperature, fO2=fO2, **kwargs
        )
        concentration = concentration + self._sulfate.concentration(
            fugacity, temperature=temperature, fO2=fO2, **kwargs
        )

        return concentration


S2_basalt_boulliung23: Solubility = _S2_basalt_boulliung23()
"""Sulfur in basalt due to sulfide and sulfate dissolution :cite:p:`BW22,BW23corr,BW23`"""


class _S2_sulfate_trachybasalt_boulliung23(Solubility):
    """Sulfur as sulfate SO4^2-/S^6+ in trachybasalt :cite:p:`BW22,BW23corr`

    Using the first equation in the abstract of :cite:t:`BW22` and the corrected expression for
    sulfate capacity (C_S6+) in :cite:t:`BW23corr`. Composition for trachybasalt from
    :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K for silicate melts
    equilibrated with Air/SO2 mixtures.
    """

    @override
    @eqx.filter_jit
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        del kwargs
        logcs: Array = -12.948 + (32446.366 / jnp.asarray(temperature))
        logso4_wtp: Array = logcs + (0.5 * jnp.log10(fugacity)) + (1.5 * jnp.log10(fO2))
        so4_wtp: Array = jnp.power(10, logso4_wtp)
        s_wtp: Array = so4_wtp * (32.065 / 96.06)
        ppmw: Array = s_wtp * unit_conversion.percent_to_ppm

        return ppmw


S2_sulfate_trachybasalt_boulliung23: Solubility = _S2_sulfate_trachybasalt_boulliung23()
"""Sulfur as sulfate SO4^2-/S^6+ in trachybasalt :cite:p:`BW22,BW23corr`

Using the first equation in the abstract of :cite:t:`BW22` and the corrected expression for
sulfate capacity (C_S6+) in :cite:t:`BW23corr`. Composition for trachybasalt from
:cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K for silicate melts
equilibrated with Air/SO2 mixtures.
"""


class _S2_sulfide_trachybasalt_boulliung23(Solubility):
    """Sulfur as sulfide (S^2-) in trachybasalt :cite:p:`BW23`

    Using expressions in the abstract for S wt.% and sulfide capacity (C_S2-). Composition
    for trachybasalt from :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K in a
    controlled CO-CO2-SO2 atmosphere fO2 conditions were greater than 1 log unit below FMQ.
    """

    @override
    @eqx.filter_jit
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        del kwargs
        logcs: Array = 0.225 - (7842.5 / jnp.asarray(temperature))
        logs_wtp: Array = logcs - (0.5 * (jnp.log10(fO2) - jnp.log10(fugacity)))
        s_wtp: Array = jnp.power(10, logs_wtp)
        ppmw: Array = s_wtp * unit_conversion.percent_to_ppm

        return ppmw


S2_sulfide_trachybasalt_boulliung23: Solubility = _S2_sulfide_trachybasalt_boulliung23()
"""Sulfur as sulfide (S^2-) in trachybasalt :cite:p:`BW23`

Using expressions in the abstract for S wt.% and sulfide capacity (C_S2-). Composition
for trachybasalt from :cite:t:`BW22{Table 1}`. Experiments conducted at 1 atm, 1473-1773 K in a
controlled CO-CO2-SO2 atmosphere fO2 conditions were greater than 1 log unit below FMQ.
"""


class _S2_trachybasalt_boulliung23(Solubility):
    """Sulfur in trachybasalt by sulfide and sulfate dissolution :cite:p:`BW22,BW23corr,BW23`"""

    _sulfide: Solubility = eqx.field(init=False)
    _sulfate: Solubility = eqx.field(init=False)

    def __post_init__(self):
        self._sulfide = S2_sulfide_trachybasalt_boulliung23
        self._sulfate = S2_sulfate_trachybasalt_boulliung23

    @override
    @eqx.filter_jit
    def concentration(
        self, fugacity: ArrayLike, *, temperature: ArrayLike, fO2: ArrayLike, **kwargs
    ) -> Array:
        concentration: ArrayLike = self._sulfide.concentration(
            fugacity, temperature=temperature, fO2=fO2, **kwargs
        )
        concentration = concentration + self._sulfate.concentration(
            fugacity, temperature=temperature, fO2=fO2, **kwargs
        )

        return concentration


S2_trachybasalt_boulliung23: Solubility = _S2_trachybasalt_boulliung23()
"""Sulfur in trachybasalt by sulfide and sulfate dissolution :cite:p:`BW22,BW23corr,BW23`"""



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Real gas EOS from :cite:`HWZ58`"""

import logging

import equinox as eqx
import jax.numpy as jnp
import optimistix as optx
from jax import Array
from jax.typing import ArrayLike

from atmodeller.constants import ATMOSPHERE, GAS_CONSTANT_BAR
from atmodeller.eos import ABSOLUTE_TOLERANCE, RELATIVE_TOLERANCE, THROW
from atmodeller.eos._aggregators import CombinedRealGas
from atmodeller.eos.core import RealGas
from atmodeller.utilities import ExperimentalCalibration, OptxSolver, unit_conversion

try:
    from typing import override  # type: ignore valid for Python 3.12+
except ImportError:
    from typing_extensions import override  # Python 3.11 and earlier

logger: logging.Logger = logging.getLogger(__name__)

# Coefficients from Table I, which must be converted to the correct units scheme (SI and pressure
# in bar). Using the original values in the paper also facilitates visual comparison and checking.


def volume_conversion(x: float) -> float:
    """Volume conversion for :cite:t:`HWZ58` units"""
    return x * unit_conversion.litre_to_m3


def A0_conversion(x: float) -> float:
    """:math:`PV^2` conversion for :cite:t:`HWZ58` units"""
    return x * ATMOSPHERE * unit_conversion.litre_to_m3**2


def atm2bar(x: float) -> float:
    """Atmosphere to bar conversion"""
    return unit_conversion.atmosphere_to_bar * x


class BeattieBridgeman(RealGas):
    r"""Beattie-Bridgeman equation :cite:p:`HWZ58{Equation 1}`

    .. math::

        PV^2 = RT\left(1-\frac{c}{VT^3}\right)\left(V+B_0-\frac{bB_0}{V}\right)
        - A_0\left(1-\frac{a}{V}\right)

    Args:
        A0: A0 empirical constant
        a: a empirical constant
        B0: B0 empirical constant
        b: b empirical constant
        c: c empirical constant
    """

    A0: float
    """A0 empirical constant"""
    a: float
    """a empirical constant"""
    B0: float
    """B0 empirical constant"""
    b: float
    """b empirical constant"""
    c: float
    """c empirical constant"""

    @eqx.filter_jit
    def _objective_function(self, volume: ArrayLike, kwargs: dict[str, ArrayLike]) -> Array:
        r"""Objective function to solve for the volume :cite:p:`HWZ58{Equation 2}`

        .. math::

            PV^4 - RTV^3 - \left(RTB_0 - \frac{Rc}{T^2}-A_0\right)V^2
            +\left(RTbB_0+\frac{RcB_0}{T^2}-aA_0\right)V - \frac{RcbB_0}{T^2}=0

        Args:
            volume: Volume in :math:`\mathrm{m}^3\ \mathrm{mol}^{-1}`
            kwargs: Dictionary with other required parameters

        Returns:
            Residual of the objective function
        """
        temperature: ArrayLike = kwargs["temperature"]
        pressure: ArrayLike = kwargs["pressure"]

        # jax.debug.print("volume = {out}", out=volume)
        # jax.debug.print("temperature = {out}", out=temperature)
        # jax.debug.print("pressure = {out}", out=pressure)

        coeff0: Array = 1 / jnp.square(temperature) * -GAS_CONSTANT_BAR * self.c * self.b * self.B0
        coeff1: Array = (
            1 / jnp.square(temperature) * GAS_CONSTANT_BAR * self.c * self.B0
            + GAS_CONSTANT_BAR * temperature * self.b * self.B0
            - self.a * self.A0
        )
        coeff2: Array = (
            1 / jnp.square(temperature) * GAS_CONSTANT_BAR * self.c
            - GAS_CONSTANT_BAR * temperature * self.B0
            + self.A0
        )
        coeff3: ArrayLike = -GAS_CONSTANT_BAR * temperature

        residual: Array = (
            coeff0
            + coeff1 * volume
            + coeff2 * jnp.power(volume, 2)
            + coeff3 * jnp.power(volume, 3)
            + pressure * jnp.power(volume, 4)
        )

        return residual

    @override
    @eqx.filter_jit
    def log_fugacity(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """Log fugacity :cite:p:`HWZ58{Equation 11}`.

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            Log fugacity
        """
        volume: ArrayLike = self.volume(temperature, pressure)
        log_fugacity: Array = (
            jnp.log(GAS_CONSTANT_BAR * temperature / volume)
            + (
                self.B0
                - self.c / jnp.power(temperature, 3)
                - self.A0 / (GAS_CONSTANT_BAR * temperature)
            )
            * 2
            / volume
            - (
                self.b * self.B0
                + self.c * self.B0 / jnp.power(temperature, 3)
                - self.a * self.A0 / (GAS_CONSTANT_BAR * temperature)
            )
            * 3
            / (2 * jnp.square(volume))
            + (self.c * self.b * self.B0 / jnp.power(temperature, 3))
            * 4
            / (3 * jnp.power(volume, 3))
        )

        return log_fugacity

    @override
    @eqx.filter_jit
    def volume(self, temperature: ArrayLike, pressure: ArrayLike) -> ArrayLike:
        r"""Solves the BB equation numerically to compute the volume.

        :cite:t:`HWZ58` doesn't say which root to take, but one real root is very small and the
        maximum real root gives a volume that agrees with the tabulated compressibility factor
        for all species.

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            Volume in :math:`\mathrm{m}^3\ \mathrm{mol}^{-1}`
        """
        initial_volume: ArrayLike = GAS_CONSTANT_BAR * temperature / pressure
        kwargs: dict[str, ArrayLike] = {"temperature": temperature, "pressure": pressure}

        solver: OptxSolver = optx.Newton(rtol=RELATIVE_TOLERANCE, atol=ABSOLUTE_TOLERANCE)
        sol = optx.root_find(
            self._objective_function, solver, initial_volume, args=kwargs, throw=THROW
        )
        volume = sol.value
        # jax.debug.print("volume = {out}", out=volume)

        return volume

    @override
    @eqx.filter_jit
    def volume_integral(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        return self.log_fugacity(temperature, pressure) * GAS_CONSTANT_BAR * temperature


pressure_min: float = atm2bar(0.1)
"""Minimum pressure for :cite:t:`HWZ58`"""

H2_beattie_holley58: RealGas = BeattieBridgeman(
    A0=A0_conversion(0.1975),
    a=volume_conversion(-0.00506),
    B0=volume_conversion(0.02096),
    b=volume_conversion(-0.04359),
    c=volume_conversion(0.0504e4),
)
"""H2 Beattie-Bridgeman :cite:p:`HWZ58`"""
H2_beattie_holley58_bounded: RealGas = CombinedRealGas.create(
    [H2_beattie_holley58],
    [ExperimentalCalibration(30, 1000, pressure_min, atm2bar(1000))],
)
"""H2 Beattie-Bridgeman bounded :cite:p:`HWZ58`"""

N2_beattie_holley58: RealGas = BeattieBridgeman(
    A0=A0_conversion(1.3445),
    a=volume_conversion(0.02617),
    B0=volume_conversion(0.05046),
    b=volume_conversion(-0.00691),
    c=volume_conversion(4.2e4),
)
"""N2 Beattie-Bridgeman :cite:p:`HWZ58`"""
N2_beattie_holley58_bounded: RealGas = CombinedRealGas.create(
    [N2_beattie_holley58], [ExperimentalCalibration(70, 1000, pressure_min, atm2bar(1000))]
)
"""N2 Beattie-Bridgeman bounded :cite:p:`HWZ58`"""

O2_beattie_holley58: RealGas = BeattieBridgeman(
    A0=A0_conversion(1.4911),
    a=volume_conversion(0.02562),
    B0=volume_conversion(0.04624),
    b=volume_conversion(0.004208),
    c=volume_conversion(4.8e4),
)
"""O2 Beattie-Bridgeman :cite:p:`HWZ58`"""
O2_beattie_holley58_bounded: RealGas = CombinedRealGas.create(
    [O2_beattie_holley58], [ExperimentalCalibration(100, 1000, pressure_min, atm2bar(1000))]
)
"""O2 Beattie-Bridgeman bounded :cite:p:`HWZ58`"""

CO2_beattie_holley58: RealGas = BeattieBridgeman(
    A0=A0_conversion(5.0065),
    a=volume_conversion(0.07132),
    B0=volume_conversion(0.10476),
    b=volume_conversion(0.07235),
    c=volume_conversion(66e4),
)
"""CO2 Beattie-Bridgeman :cite:p:`HWZ58`"""
CO2_beattie_holley58_bounded: RealGas = CombinedRealGas.create(
    [CO2_beattie_holley58],
    [ExperimentalCalibration(200, 1000, pressure_min, atm2bar(1000))],
)
"""CO2 Beattie-Bridgeman bounded :cite:p:`HWZ58`"""

NH3_beattie_holley58: RealGas = BeattieBridgeman(
    A0=A0_conversion(2.3930),
    a=volume_conversion(0.17031),
    B0=volume_conversion(0.03415),
    b=volume_conversion(0.19112),
    c=volume_conversion(476.87e4),
)
"""NH3 Beattie-Bridgeman :cite:p:`HWZ58`"""
NH3_beattie_holley58_bounded: RealGas = CombinedRealGas.create(
    [NH3_beattie_holley58], [ExperimentalCalibration(300, 1000, pressure_min, atm2bar(500))]
)
"""NH3 Beattie-Bridgeman bounded :cite:p:`HWZ58`"""

CH4_beattie_holley58: RealGas = BeattieBridgeman(
    A0=A0_conversion(2.2769),
    a=volume_conversion(0.01855),
    B0=volume_conversion(0.05587),
    b=volume_conversion(-0.01587),
    c=volume_conversion(12.83e4),
)
"""CH4 Beattie-Bridgeman :cite:p:`HWZ58`"""
CH4_beattie_holley58_bounded: RealGas = CombinedRealGas.create(
    [CH4_beattie_holley58],
    [ExperimentalCalibration(100, 1000, pressure_min, atm2bar(1000))],
)
"""CH4 Beattie-Bridgeman bounded :cite:p:`HWZ58`"""

He_beattie_holley58: RealGas = BeattieBridgeman(
    A0=A0_conversion(0.0216),
    a=volume_conversion(0.05984),
    B0=volume_conversion(0.01400),
    b=0,
    c=volume_conversion(0.004e4),
)
"""He Beattie-Bridgeman :cite:p:`HWZ58`"""
He_beattie_holley58_bounded: RealGas = CombinedRealGas.create(
    [He_beattie_holley58],
    [ExperimentalCalibration(10, 1000, pressure_min, atm2bar(1000))],
)
"""He Beattie-Bridgeman bounded :cite:p:`HWZ58`"""


def get_holley_eos_models() -> dict[str, RealGas]:
    """Gets a dictionary of Holley EOS models

    The naming convention is as follows:
        [species]_[eos model]_[citation]

    Returns:
        Dictionary of EOS models
    """
    eos_models: dict[str, RealGas] = {}
    eos_models["CH4_beattie_holley58"] = CH4_beattie_holley58_bounded
    eos_models["CO2_beattie_holley58"] = CO2_beattie_holley58_bounded
    eos_models["H2_beattie_holley58"] = H2_beattie_holley58_bounded
    eos_models["He_beattie_holley58"] = He_beattie_holley58_bounded
    eos_models["N2_beattie_holley58"] = N2_beattie_holley58_bounded
    eos_models["NH3_beattie_holley58"] = NH3_beattie_holley58_bounded
    eos_models["O2_beattie_holley58"] = O2_beattie_holley58_bounded

    return eos_models



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Package level variables and initialises the package logger"""

from __future__ import annotations

__version__: str = "0.4.0"

import logging

import jax
import jax.numpy as jnp
import numpy as np

jax.config.update("jax_enable_x64", True)
jnp.set_printoptions(precision=15)  # For better clarity in printed output
print("Atmodeller initialized with double precision (float64)")

# For debugging
# jax.config.update("jax_debug_nans", True)
# jax.config.update("jax_debug_infs", False)
# jax.config.update("jax_disable_jit", True)
# jax.config.update("jax_log_compiles", True)

# Thermodynamic standard state
ENTHALPY_REFERENCE: float = 298.15
"""Enthalpy reference temperature in K"""
PRESSURE_REFERENCE: float = 1.0
"""Standard state pressure in bar"""

# Initial solution guess
INITIAL_LOG_NUMBER_DENSITY: float = 50.0
"""Initial log number density

Empiricially determined. This value is mid-range for Earth-like planets.
"""
INITIAL_LOG_STABILITY: float = -140.0
"""Initial log stability

Empirically determined. Preliminary testing seems to reveal that starting from a large negative 
value improves the performance of the solver. This asserts that all species are stable.
"""

# Maximum x for which exp(x) is finite in 64-bit precision (to prevent overflow)
max_exp_input = np.log(np.finfo(np.float64).max)
# Minimum x for which exp(x) is non-zero in 64-bit precision
min_exp_input = np.log(jnp.finfo(np.float64).tiny)

# Lower and upper bounds on the hypercube which contains the root
LOG_NUMBER_DENSITY_LOWER: float = -170.0
"""Lower log number density for a species

At 3000 K this corresponds to 3.17E-77 bar and at 298 K this corresponds to 3.16E-78 bar.
"""
LOG_NUMBER_DENSITY_UPPER: float = 80.0
"""Upper log number density for a species

TODO: The summary values below are for a value of 70.0, but it has been increased to 80 since this
cut-off also affects condensates, which can extend to large values. The hypercube specification
can take a pytree structurs, so a more sophisticated approach can be implemented where gas species
and condensate are treated differently.

At 3000 K this corresponds to 1041881 bar (104 GPa) and at 298 K this corresponds to 103494 bar
(10.3 GPa).
"""
LOG_STABILITY_LOWER: float = -700.0  # basically the same as min_exp_input
"""Lower stability for a species

Derived to ensure that the exponential function exp(x) does not underflow to zero
"""
LOG_STABILITY_UPPER: float = 35.0
"""Upper stability for a species

Empirically determined.
"""
TAU: float = 1.0e-25
"""Tau scaling factor for species stability :cite:p:`LKK16`.

Tau effectively controls the minimum non-zero number density of unstable species. Formally, it
defines the number density of an unstable pure condensate with an activity of 1/e, which
corresponds to a log stability of zero.

This value is typically appropriate for condensate stability only, but if you additionally apply 
stability criteria to gas species you should reduce this value, maybe as low as 1e-60 to 1e-72 if
you want to ensure you do not truncated O2 at low temperatures. Hence you can override this default
using an argument to :class:`atmodeller.classes.InteriorAtmosphere`.
"""

# Create the package logger.
# https://docs.python.org/3/howto/logging.html#library-config
logger: logging.Logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def complex_formatter() -> logging.Formatter:
    """Complex formatter"""
    fmt: str = "[%(asctime)s - %(name)-30s - %(lineno)03d - %(levelname)-9s - %(funcName)s()]"
    fmt += " - %(message)s"
    datefmt: str = "%Y-%m-%d %H:%M:%S"
    formatter: logging.Formatter = logging.Formatter(fmt, datefmt=datefmt)

    return formatter


def simple_formatter() -> logging.Formatter:
    """Simple formatter for logging

    Returns:
        Formatter for logging
    """
    fmt: str = "[%(asctime)s - %(name)-30s - %(levelname)-9s] - %(message)s"
    datefmt: str = "%H:%M:%S"
    formatter: logging.Formatter = logging.Formatter(fmt, datefmt=datefmt)

    return formatter


def debug_logger() -> logging.Logger:
    """Sets up debug logging to the console.

    Returns:
        A logger
    """
    package_logger: logging.Logger = logging.getLogger(__name__)
    package_logger.setLevel(logging.DEBUG)
    package_logger.handlers = []
    console_handler: logging.Handler = logging.StreamHandler()
    console_formatter: logging.Formatter = simple_formatter()
    console_handler.setFormatter(console_formatter)
    logger.addHandler(console_handler)

    return package_logger


def debug_file_logger() -> logging.Logger:
    """Sets up info logging to the console and debug logging to a file.

    Returns:
        A logger
    """
    # Console logger
    package_logger: logging.Logger = logging.getLogger(__name__)
    package_logger.setLevel(logging.DEBUG)
    package_logger.handlers = []
    console_handler: logging.Handler = logging.StreamHandler()
    console_formatter: logging.Formatter = simple_formatter()
    console_handler.setFormatter(console_formatter)
    console_handler.setLevel(logging.INFO)
    package_logger.addHandler(console_handler)
    # File logger
    file_handler: logging.Handler = logging.FileHandler(f"{__package__}.log")
    file_formatter: logging.Formatter = complex_formatter()
    file_handler.setFormatter(file_formatter)
    file_handler.setLevel(logging.DEBUG)
    package_logger.addHandler(file_handler)

    return package_logger


from atmodeller.classes import InteriorAtmosphere  # noqa: E402, F401
from atmodeller.containers import (  # noqa: E402, F401
    Planet,
    SolverParameters,
    Species,
    SpeciesCollection,
)
from atmodeller.utilities import (  # noqa: E402, F401
    bulk_silicate_earth_abundances,
    earth_oceans_to_hydrogen_mass,
)



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""JAX-related functionality for solving the system of equations"""

from __future__ import annotations

from typing import Any, Callable

import equinox as eqx
import jax
import jax.numpy as jnp
import optimistix as optx
from jax import Array, lax
from jax.scipy.special import logsumexp
from jax.typing import ArrayLike

from atmodeller.constants import AVOGADRO, BOLTZMANN_CONSTANT_BAR, GAS_CONSTANT
from atmodeller.containers import (
    FixedParameters,
    FugacityConstraints,
    MassConstraints,
    Planet,
    SolverParameters,
    SpeciesCollection,
    TracedParameters,
)
from atmodeller.utilities import (
    get_log_number_density_from_log_pressure,
    safe_exp,
    unit_conversion,
)


@eqx.filter_jit
def solve(
    solution_array: Array,
    traced_parameters: TracedParameters,
    fixed_parameters: FixedParameters,
    solver_parameters: SolverParameters,
    options: dict[str, Any],
) -> tuple[Array, Array, Array]:
    """Solves the system of non-linear equations

    Args:
        solution_array: Solution array
        traced_parameters: Traced parameters
        fixed_parameters: Fixed parameters
        solver_parameters: Solver parameters
        options: Options for root find

    Returns:
        The solution array, the status of the solver
    """
    sol: optx.Solution = optx.root_find(
        objective_function,
        solver_parameters.solver_instance,
        solution_array,
        args={
            "traced_parameters": traced_parameters,
            "fixed_parameters": fixed_parameters,
        },
        throw=solver_parameters.throw,
        max_steps=solver_parameters.max_steps,
        options=options,
    )

    # jax.debug.print("Optimistix success. Number of steps = {out}", out=sol.stats["num_steps"])
    solver_steps: Array = sol.stats["num_steps"]
    solver_status: Array = sol.result == optx.RESULTS.successful

    return sol.value, solver_status, solver_steps


@eqx.filter_jit
def repeat_solver(
    solver_fn: Callable,
    base_initial_solution: Array,
    initial_solution: Array,
    initial_status: Array,
    initial_steps: Array,
    traced_parameters: TracedParameters,
    multistart_perturbation: float,
    max_attempts: int,
    key: Array,
) -> tuple[int, Array, Array, Array]:
    """Repeat solver until a solution is found

    Args:
        solver_fn: Solver function with pre-bound fixed configuration
        base_initial_solution: Base initial solution to perturb if necessary
        initial_solution: Initial solution after first solve
        initial_status: Initial status after first solve
        initial_steps: Initial steps after first solve
        traced_parameters: Traced parameters
        multistart_perturbation: Multistart perturbation
        max_attempts: Maximum attempts
        key: Random key
    """

    def body_fn(state: tuple) -> tuple[int, Array, Array, Array, Array, Array]:
        """Perform one iteration of the solver retry loop

        Args:
            state: Tuple containing:
                i: Current attempt index
                key: PRNG key for random number generation
                solution: Current solution array
                status: Boolean array indicating successful solutions
                steps: Step count or similar solver output
                base_initial_solution: Unperturbed base initial solution

        Returns:
            Updated state tuple with incremented attempt index, potentially perturbed
            solutions, updated status, updated steps, and unmodified base initial solution
        """
        i, key, solution, status, _, base_initial_solution = state

        failed_mask: Array = ~status
        key, subkey = jax.random.split(key)

        # Implements a simple perturbation of the base initial solution, but something more
        # sophisticated could be implemented, such as training a network or using a regressor
        # to inform the next guess of the models that failed from the ones that succeeded.
        perturb_shape: tuple[int, int] = (solution.shape[0], solution.shape[1])
        raw_perturb: Array = jax.random.uniform(
            subkey, shape=perturb_shape, minval=-1.0, maxval=1.0
        )
        perturbations: Array = jnp.where(
            failed_mask[:, None],
            multistart_perturbation * raw_perturb,
            jnp.zeros_like(solution),
        )
        new_initial_solution = jnp.where(
            failed_mask[:, None], base_initial_solution + perturbations, solution
        )

        new_solution, new_status, new_steps = solver_fn(new_initial_solution, traced_parameters)

        return (i + 1, key, new_solution, new_status, new_steps, base_initial_solution)

    def cond_fn(state: tuple) -> Array:
        """Check if the solver should continue retrying

        Args:
            state: Tuple containing:
                i: Current attempt index
                _: Unused (PRNG key)
                _: Unused (solution)
                status: Boolean array indicating success of each solution
                _: Unused (steps)
                _: Unused (base initial solution)

        Returns:
            A boolean array indicating whether retries should continue (True if
            any solution failed and attempts are still available)
        """
        i, _, _, status, _, _ = state

        return jnp.logical_and(i < max_attempts, jnp.any(~status))

    initial_state: tuple = (
        1,  # A first solve has already been attempted before repeat_solver is called
        key,
        initial_solution,
        initial_status,
        initial_steps,
        base_initial_solution,
    )
    final_i, _, final_solution, final_status, final_steps, _ = lax.while_loop(
        cond_fn, body_fn, initial_state
    )

    return final_i, final_solution, final_status, final_steps


@eqx.filter_jit
def get_min_log_elemental_abundance_per_species(
    formula_matrix: Array, mass_constraints: MassConstraints
) -> Array:
    """For each species, find the elemental mass constraint with the lowest abundance.

    If species are present for which there are no elemental mass constraints (for example,
    oxygen if a fugacity constraint is applied instead) the maximum of all elements will be
    returned for that particular species. However, the return array from this function is
    subsequently filtered depending on whether a stability calculation is required for each
    species.

    # TODO: This could be precomputed rather than calculated within the residual calculation.

    Args:
        formula_matrix: Formula matrix
        mass_constraints: Mass constraints

    Returns:
        A vector of the minimum log elemental abundance for each species
    """
    # Create the binary mask where formula_matrix != 0 (1 where element is present in species)
    mask: Array = (formula_matrix != 0).astype(jnp.int_)
    # jax.debug.print("formula_matrix = {out}", out=formula_matrix)
    # jax.debug.print("mask = {out}", out=mask)

    # Convert to column vector to align for element-wise multiplication
    log_abundance: Array = jnp.expand_dims(
        jnp.array(list(mass_constraints.log_abundance.values())), axis=1
    )
    # jax.debug.print("log_abundance = {out}", out=log_abundance)

    # Element-wise multiplication (broadcasted correctly)
    masked_abundance: Array = mask * log_abundance  # Shape: (n_elements, n_species)
    # jax.debug.print("masked_abundance = {out}", out=masked_abundance)

    # Replace zeros (or masked-out values) with the max value before taking the min
    masked_abundance = jnp.where(mask != 0, masked_abundance, jnp.max(log_abundance))
    # jax.debug.print("masked_abundance = {out}", out=masked_abundance)

    # Find the minimum log abundance per species
    min_abundance_per_species: Array = jnp.min(masked_abundance, axis=0)
    # jax.debug.print("min_abundance_per_species = {out}", out=min_abundance_per_species)

    return min_abundance_per_species


@eqx.filter_jit
def objective_function(solution: Array, kwargs: dict) -> Array:
    """Residual of the reaction network and mass balance

    Args:
        solution: Solution array
        kwargs: Dictionary of pytrees required to compute the residual

    Returns:
        Residual
    """
    traced_parameters: TracedParameters = kwargs["traced_parameters"]
    fixed_parameters: FixedParameters = kwargs["fixed_parameters"]
    planet: Planet = traced_parameters.planet
    temperature: ArrayLike = planet.temperature
    fugacity_constraints: FugacityConstraints = traced_parameters.fugacity_constraints
    mass_constraints: MassConstraints = traced_parameters.mass_constraints
    gas_species_indices: Array = fixed_parameters.gas_species_indices

    reaction_matrix: Array = jnp.array(fixed_parameters.reaction_matrix)
    reaction_stability_matrix: Array = jnp.array(fixed_parameters.reaction_stability_matrix)
    stability_species_indices: Array = fixed_parameters.stability_species_indices
    fugacity_matrix: Array = fixed_parameters.fugacity_matrix
    # We only need the formula matrix for elements with mass constraints
    formula_matrix_constraints: Array = jnp.array(fixed_parameters.formula_matrix_constraints)
    # jax.debug.print("Starting new residual evaluation")

    # Species
    if stability_species_indices.size > 0:
        split_idx: int = -len(stability_species_indices)
        log_number_density: Array = solution[:split_idx]
        log_stability: Array = solution[split_idx:]
        # jax.debug.print("log_stability = {out}", out=log_stability)
    else:
        log_number_density = solution

    # jax.debug.print("log_number_density = {out}", out=log_number_density)

    # Stability
    # The reaction_stability_matrix has zero entries so its OK to pad with any value for
    # species without stability since the result is always zero
    log_stability_padded: Array = jnp.zeros_like(log_number_density)
    if stability_species_indices.size > 0:
        log_stability_padded = log_stability_padded.at[stability_species_indices].set(
            log_stability  # type: ignore since log_stability is known
        )
    # jax.debug.print("log_stability_padded = {out}", out=log_stability_padded)

    log_activity: Array = get_log_activity(traced_parameters, fixed_parameters, log_number_density)
    # jax.debug.print("log_activity = {out}", out=log_activity)

    # Based on the definition of the reaction constant we need to convert gas activities
    # (fugacities) from bar to effective number density
    gas_species_mask: Array = jnp.zeros_like(log_activity, dtype=bool)
    gas_species_mask = gas_species_mask.at[gas_species_indices].set(True)
    # jax.debug.print("gas_species_mask = {out}", out=gas_species_mask)

    log_activity_number_density: Array = get_log_number_density_from_log_pressure(
        log_activity, temperature
    )
    log_activity_number_density = jnp.where(
        gas_species_mask, log_activity_number_density, log_activity
    )
    # jax.debug.print("log_activity_number_density = {out}", out=log_activity_number_density)

    # Bulk atmosphere
    log_volume: Array = get_atmosphere_log_volume(fixed_parameters, log_number_density, planet)

    residual: Array = jnp.array([])

    # Reaction network residual
    if reaction_matrix.size > 0:
        log_reaction_equilibrium_constant: Array = get_log_reaction_equilibrium_constant(
            fixed_parameters, temperature
        )
        reaction_residual: Array = (
            reaction_matrix.dot(log_activity_number_density) - log_reaction_equilibrium_constant
        )
        # jax.debug.print("reaction_residual before stability = {out}", out=reaction_residual)

        reaction_residual = reaction_residual - reaction_stability_matrix.dot(
            safe_exp(log_stability_padded)
        )
        # jax.debug.print("reaction_residual after stability = {out}", out=reaction_residual)

        residual = jnp.concatenate([residual, reaction_residual])

    # Fugacity constraints residual
    if fugacity_matrix.size > 0:
        fugacity_species_indices: Array = jnp.array(fixed_parameters.fugacity_species_indices)
        # jax.debug.print("fugacity_species_indices = {out}", out=fugacity_species_indices)
        fugacity_log_activity_number_density: Array = jnp.take(
            log_activity_number_density, fugacity_species_indices
        )
        # jax.debug.print(
        #    "fugacity_log_activity_number_density = {out}",
        #    out=fugacity_log_activity_number_density,
        # )
        # jax.debug.print("fugacity_matrix = {out}", out=fugacity_matrix)
        # TODO: Check if there is a need to consider the stability of the species for which the
        # fugacity constraint is being applied.
        fugacity_residual: Array = fugacity_matrix.dot(fugacity_log_activity_number_density)
        # jax.debug.print("fugacity_residual = {out}", out=fugacity_residual)
        total_pressure: Array = get_total_pressure(
            fixed_parameters, log_number_density, temperature
        )
        fugacity_residual = fugacity_residual - fugacity_constraints.log_number_density(
            temperature, total_pressure
        )
        # jax.debug.print("fugacity_residual = {out}", out=fugacity_residual)
        residual = jnp.concatenate([residual, fugacity_residual])

    # Elemental mass balance residual
    if formula_matrix_constraints.size > 0:
        # Number density of elements in the gas or condensed phase
        element_density: Array = get_element_density(
            formula_matrix_constraints, log_number_density
        )
        # jax.debug.print("element_density = {out}", out=element_density)
        element_melt_density: Array = get_element_density_in_melt(
            traced_parameters,
            fixed_parameters,
            formula_matrix_constraints,
            log_number_density,
            log_activity,
            log_volume,
        )
        # jax.debug.print("element_melt_density = {out}", out=element_melt_density)

        # Relative mass error, computed in log-space for numerical stability
        element_density_total: Array = element_density + element_melt_density
        log_element_density_total: Array = jnp.log(element_density_total)
        log_target_density: Array = mass_constraints.log_number_density(log_volume)
        mass_residual: Array = safe_exp(log_element_density_total - log_target_density) - 1
        # jax.debug.print("element_density = {out}", out=mass_residual)

        residual = jnp.concatenate([residual, mass_residual])

        # Stability residual
        if stability_species_indices.size > 0:
            log_min_number_density: Array = (
                get_min_log_elemental_abundance_per_species(
                    formula_matrix_constraints, mass_constraints
                )
                - log_volume
                + jnp.log(fixed_parameters.tau)
            )
            log_min_number_density = jnp.take(log_min_number_density, stability_species_indices)

            log_number_density_stability: Array = jnp.take(
                log_number_density, stability_species_indices
            )
            stability_residual: Array = (
                log_number_density_stability + log_stability - log_min_number_density  # type: ignore
            )
            # jax.debug.print("stability_residual = {out}", out=stability_residual)

            residual = jnp.concatenate([residual, stability_residual])

    # jax.debug.print("residual = {out}", out=residual)

    return residual


@eqx.filter_jit
def get_atmosphere_log_molar_mass(
    fixed_parameters: FixedParameters, log_number_density: Array
) -> Array:
    """Gets log molar mass of the atmosphere

    Args:
        fixed_parameters: Fixed parameters
        log_number_density: Log number density

    Returns:
        Log molar mass of the atmosphere
    """
    gas_log_number_density: Array = get_gas_species_data(fixed_parameters, log_number_density)
    gas_molar_mass: Array = get_gas_species_data(
        fixed_parameters, jnp.array(fixed_parameters.molar_masses)
    )
    molar_mass: Array = logsumexp(gas_log_number_density, b=gas_molar_mass) - logsumexp(
        gas_log_number_density
    )

    return molar_mass


@eqx.filter_jit
def get_atmosphere_log_volume(
    fixed_parameters: FixedParameters,
    log_number_density: Array,
    planet: Planet,
) -> Array:
    """Gets log volume of the atmosphere

    Args:
        fixed_parameters: Fixed parameters
        log_number_density: Log number density
        planet: Planet

    Returns:
        Log volume of the atmosphere
    """
    log_volume: Array = (
        jnp.log(GAS_CONSTANT)
        + jnp.log(planet.temperature)
        - get_atmosphere_log_molar_mass(fixed_parameters, log_number_density)
        + jnp.log(planet.surface_area)
        - jnp.log(planet.surface_gravity)
    )

    return log_volume


@eqx.filter_jit
def get_total_pressure(
    fixed_parameters: FixedParameters, log_number_density: Array, temperature: ArrayLike
) -> Array:
    """Gets total pressure

    Args:
        fixed_parameters: Fixed parameters
        log_number_density: Log number density
        temperature: Temperature

    Returns:
        Total pressure
    """
    gas_species_indices: Array = jnp.array(fixed_parameters.gas_species_indices)
    pressure: Array = get_pressure_from_log_number_density(log_number_density, temperature)
    gas_pressure: Array = jnp.take(pressure, gas_species_indices)
    # jax.debug.print("gas_pressure = {out}", out=gas_pressure)

    return jnp.sum(gas_pressure)


@eqx.filter_jit
def get_element_density(formula_matrix: Array, log_number_density: Array) -> Array:
    """Number density of elements in the gas or condensed phase

    Args:
        formula_matrix: Formula matrix
        log_number_density: Log number density

    Returns:
        Number density of elements in the gas or condensed phase
    """
    element_density: Array = formula_matrix.dot(safe_exp(log_number_density))

    return element_density


@eqx.filter_jit
def get_element_density_in_melt(
    traced_parameters: TracedParameters,
    fixed_parameters: FixedParameters,
    formula_matrix: Array,
    log_number_density: Array,
    log_activity: Array,
    log_volume: Array,
) -> Array:
    """Gets the number density of elements dissolved in melt due to species solubility

    Args:
        traced_parameters: Traced parameters
        fixed_parameters: Fixed parameters
        formula_matrix: Formula matrix
        log_number_density: Log number density
        log_activity: Log activity
        log_volume: Log volume of the atmosphere

    Returns:
        Number density of elements dissolved in melt
    """
    species_melt_density: Array = get_species_density_in_melt(
        traced_parameters, fixed_parameters, log_number_density, log_activity, log_volume
    )
    element_melt_density: Array = formula_matrix.dot(species_melt_density)

    return element_melt_density


@eqx.filter_jit
def get_gas_species_data(fixed_parameters: FixedParameters, some_array: ArrayLike) -> Array:
    """Gets the gas species data from an array

    Args:
        fixed_parameters: Fixed parameters
        some_array: Some array to extract gas species data from

    Returns:
        An array with just the gas species data from `some_array`
    """
    gas_species_indices: Array = jnp.array(fixed_parameters.gas_species_indices)
    gas_data: Array = jnp.take(some_array, gas_species_indices)

    return gas_data


@eqx.filter_jit
def get_log_activity(
    traced_parameters: TracedParameters,
    fixed_parameters: FixedParameters,
    log_number_density: Array,
) -> Array:
    """Gets the log activity

    Args:
        traced_parameters: Traced parameters
        fixed_parameters: Fixed parameters
        log_number_density: Log number density

    Returns:
        Log activity
    """
    planet: Planet = traced_parameters.planet
    temperature: ArrayLike = planet.temperature
    species: SpeciesCollection = fixed_parameters.species
    total_pressure: Array = get_total_pressure(fixed_parameters, log_number_density, temperature)
    # jax.debug.print("total_pressure = {out}", out=total_pressure)

    activity_funcs: list[Callable] = [species_.activity.log_activity for species_ in species]

    def apply_activity_function(index: ArrayLike):
        return lax.switch(
            index,
            activity_funcs,
            temperature,
            total_pressure,
        )

    vmap_apply_function: Callable = eqx.filter_vmap(apply_activity_function, in_axes=(0,))
    indices: Array = jnp.arange(len(species))
    log_activity_pure_species: Array = vmap_apply_function(indices)
    # jax.debug.print("log_activity_pure_species = {out}", out=log_activity_pure_species)
    log_activity = get_log_activity_ideal_mixing(
        fixed_parameters, log_number_density, log_activity_pure_species
    )
    # jax.debug.print("log_activity = {out}", out=log_activity)

    return log_activity


@eqx.filter_jit
def get_log_activity_ideal_mixing(
    fixed_parameters: FixedParameters, log_number_density: Array, log_activity_pure_species: Array
) -> Array:
    """Gets the log activity of species in the atmosphere assuming an ideal mixture

    Args:
        fixed_parameters: Fixed parameters
        log_number_density: Log number density
        log_activity_pure_species: Log activity of the pure species

    Returns:
        Log activity of the species assuming ideal mixing in the atmosphere
    """
    gas_species_indices: Array = jnp.array(fixed_parameters.gas_species_indices)
    gas_species_mask: Array = jnp.zeros_like(log_activity_pure_species, dtype=bool)
    gas_species_mask = gas_species_mask.at[gas_species_indices].set(True)

    number_density: Array = safe_exp(log_number_density)
    gas_species_number_density: Array = jnp.where(gas_species_mask, number_density, 0)
    atmosphere_log_number_density: Array = jnp.log(jnp.sum(gas_species_number_density))

    log_activity_gas_species: Array = (
        log_activity_pure_species + log_number_density - atmosphere_log_number_density
    )
    log_activity: Array = jnp.where(
        gas_species_mask, log_activity_gas_species, log_activity_pure_species
    )

    return log_activity


@eqx.filter_jit
def get_log_pressure_from_log_number_density(
    log_number_density: ArrayLike, temperature: ArrayLike
) -> Array:
    """Gets log pressure from log number density

    Args:
        log_number_density: Log number density
        temperature: Temperature

    Returns:
        Log pressure
    """
    log_pressure: Array = (
        jnp.log(BOLTZMANN_CONSTANT_BAR) + jnp.log(temperature) + log_number_density
    )

    return log_pressure


@eqx.filter_jit
def get_log_Kp(
    species: SpeciesCollection, reaction_matrix: Array, temperature: ArrayLike
) -> Array:
    """Gets log of the equilibrium constant in terms of partial pressures

    Args:
        species: Species
        reaction_matrix: Reaction matrix
        temperature: Temperature

    Returns:
        Log of the equilibrium constant in terms of partial pressures
    """
    gibbs_list: list[ArrayLike] = []
    for species_ in species:
        gibbs: ArrayLike = species_.data.get_gibbs_over_RT(temperature)
        gibbs_list.append(gibbs)

    gibbs_jnp: Array = jnp.array(gibbs_list)
    log_Kp: Array = -1.0 * reaction_matrix.dot(gibbs_jnp)

    return log_Kp


@eqx.filter_jit
def get_log_reaction_equilibrium_constant(
    fixed_parameters: FixedParameters,
    temperature: ArrayLike,
) -> Array:
    """Gets the log equilibrium constant of the reactions

    Args:
        fixed_parameters: Fixed parameters
        temperature: Temperature

    Returns:
        Log equilibrium constant of the reactions
    """
    species: SpeciesCollection = fixed_parameters.species
    reaction_matrix: Array = jnp.array(fixed_parameters.reaction_matrix)
    gas_species_indices: Array = jnp.array(fixed_parameters.gas_species_indices)

    log_Kp: Array = get_log_Kp(species, reaction_matrix, temperature)
    # jax.debug.print("lnKp = {out}", out=lnKp)
    delta_n: Array = jnp.sum(jnp.take(reaction_matrix, gas_species_indices, axis=1), axis=1)
    # jax.debug.print("delta_n = {out}", out=delta_n)
    log_Kc: Array = log_Kp - delta_n * (jnp.log(BOLTZMANN_CONSTANT_BAR) + jnp.log(temperature))
    # jax.debug.print("log10Kc = {out}", out=log_Kc)

    return log_Kc


@eqx.filter_jit
def get_pressure_from_log_number_density(
    log_number_density: ArrayLike, temperature: ArrayLike
) -> Array:
    """Gets pressure from log number density

    Args:
        log_number_density: Log number density
        temperature: Temperature

    Returns:
        Pressure
    """
    return safe_exp(get_log_pressure_from_log_number_density(log_number_density, temperature))


@eqx.filter_jit
def get_species_density_in_melt(
    traced_parameters: TracedParameters,
    fixed_parameters: FixedParameters,
    log_number_density: Array,
    log_activity: Array,
    log_volume: Array,
) -> Array:
    """Gets the number density of species dissolved in melt due to species solubility

    Args:
        traced_parameters: Traced parameters
        fixed_parameters: Fixed parameters
        log_number_density: Log number density
        log_activity: Log activity
        log_volume: Log volume of the atmosphere

    Returns:
        Number density of species dissolved in melt
    """
    molar_masses: Array = jnp.array(fixed_parameters.molar_masses)
    planet: Planet = traced_parameters.planet

    ppmw: Array = get_species_ppmw_in_melt(
        traced_parameters, fixed_parameters, log_number_density, log_activity
    )

    species_melt_density: Array = (
        ppmw
        * unit_conversion.ppm_to_fraction
        * AVOGADRO
        * planet.melt_mass
        / (molar_masses * safe_exp(log_volume))
    )
    # jax.debug.print("species_melt_density = {out}", out=species_melt_density)

    return species_melt_density


@eqx.filter_jit
def get_species_ppmw_in_melt(
    traced_parameters: TracedParameters,
    fixed_parameters: FixedParameters,
    log_number_density: Array,
    log_activity: Array,
) -> Array:
    """Gets the ppmw of species dissolved in melt due to species solubility

    Args:
        traced_parameters: Traced parameters
        fixed_parameters: Fixed parameters
        log_number_density: Log number density
        log_activity: Log activity

    Returns:
        ppmw of species dissolved in melt
    """
    species: SpeciesCollection = fixed_parameters.species
    diatomic_oxygen_index: Array = jnp.array(fixed_parameters.diatomic_oxygen_index)
    planet: Planet = traced_parameters.planet
    temperature: ArrayLike = planet.temperature

    fugacity: Array = safe_exp(log_activity)
    total_pressure: Array = get_total_pressure(fixed_parameters, log_number_density, temperature)
    diatomic_oxygen_fugacity: Array = jnp.take(fugacity, diatomic_oxygen_index)

    # NOTE: All solubility formulations must return a JAX array to allow vmapping
    solubility_funcs: list[Callable] = [
        species_.solubility.jax_concentration for species_ in species
    ]

    def apply_solubility_function(index: ArrayLike, fugacity: ArrayLike):
        return lax.switch(
            index,
            solubility_funcs,
            fugacity,
            temperature,
            total_pressure,
            diatomic_oxygen_fugacity,
        )

    vmap_apply_function: Callable = eqx.filter_vmap(apply_solubility_function, in_axes=(0, 0))
    indices: ArrayLike = jnp.arange(len(species))
    species_ppmw: Array = vmap_apply_function(indices, fugacity)
    # jax.debug.print("ppmw = {out}", out=ppmw)

    return species_ppmw



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Real gas EOS from :cite:t:`CD21`"""

import importlib.resources
import logging
from contextlib import AbstractContextManager
from pathlib import Path
from typing import Callable, ClassVar

import equinox as eqx
import jax.numpy as jnp
import pandas as pd
from jax import Array
from jax.scipy.interpolate import RegularGridInterpolator
from jax.typing import ArrayLike
from molmass import Formula
from xmmutablemap import ImmutableMap

from atmodeller import PRESSURE_REFERENCE
from atmodeller.constants import GAS_CONSTANT_BAR
from atmodeller.eos import DATA_DIRECTORY
from atmodeller.eos._aggregators import CombinedRealGas
from atmodeller.eos.core import RealGas
from atmodeller.utilities import ExperimentalCalibration, unit_conversion

try:
    from typing import override  # type: ignore valid for Python 3.12+
except ImportError:
    from typing_extensions import override  # Python 3.11 and earlier

logger: logging.Logger = logging.getLogger(__name__)


class Chabrier(RealGas):
    r"""Chabrier EOS from :cite:t:`CD21`

    This uses rho-T-P tables to lookup density (rho).

    Args:
        log10_density_func: Spline lookup for density from :cite:t:`CD21` T-P-rho tables
        He_fraction: He fraction
        H2_molar_mass_g_mol: Molar mass of H2
        He_molar_mass_g_mol: Molar mass of He
        integration_steps: Number of integration steps
    """

    CHABRIER_DIRECTORY: ClassVar[Path] = Path("chabrier")
    """Directory of the Chabrier data within :obj:`~atmodeller.eos.data`"""
    He_fraction_map: ClassVar[ImmutableMap[str, float]] = ImmutableMap(
        {
            "TABLE_H_TP_v1": 0,
            "TABLE_HE_TP_v1": 1,
            "TABLEEOS_2021_TP_Y0275_v1": 0.275,
            "TABLEEOS_2021_TP_Y0292_v1": 0.292,
            "TABLEEOS_2021_TP_Y0297_v1": 0.297,
        }
    )
    """Mole fraction of He in the gas mixture, the other component being H2.
    
    Dictionary keys should correspond to the name of the Chabrier file.
    """
    log10_density_func: Callable
    """Spline lookup for density from :cite:t:`CD21` T-P-rho tables"""
    He_fraction: float
    """He fraction"""
    H2_molar_mass_g_mol: float
    """Molar mass of H2"""
    He_molar_mass_g_mol: float
    """Molar mass of He"""
    integration_steps: int
    """Number of integration steps"""

    @classmethod
    def create(cls, filename: Path, integration_steps: int = 100) -> RealGas:
        """Creates a Chabrier instance

        Args:
            filename: Filename of the density-T-P data
            integration_steps: Number of integration steps. Defaults to 100, which computes the
                fugacity of H2 to within 4% (relative to 1000 steps, for T from 1000 to 5000 K and
                pressure to 10 GPa), and to within 10% (relative to 1000 steps, for T from 1000 to
                5000 K and pressure to 100 GPa). Increasing the integration steps will increase the
                run time but provide better accuracy.

        Returns:
            Instance
        """
        log10_density_func: Callable = cls._get_interpolator(filename)
        He_fraction: float = cls.He_fraction_map[filename.name]
        H2_molar_mass_g_mol: float = Formula("H2").mass
        He_molar_mass_g_mol: float = Formula("He").mass

        return cls(
            log10_density_func,
            He_fraction,
            H2_molar_mass_g_mol,
            He_molar_mass_g_mol,
            integration_steps,
        )

    @classmethod
    def _get_interpolator(cls, filename: Path) -> Callable:
        """Gets spline lookup for density from :cite:t:`CD21` T-P-rho tables.

        The data tables have a slightly different organisation of the header line. But in all cases
        the first three columns contain the required data: log10 T [K], log10 P [GPa], and
        log10 rho [g/cc].

        Args:
            filename: Filename of the density-T-P data

        Returns:
            Interpolator
        """
        # Define column names for the first three columns
        T_name: str = "log T [K]"
        P_name: str = "log P [GPa]"
        rho_name: str = "log rho [g/cc]"
        column_names: list[str] = [T_name, P_name, rho_name]

        data: AbstractContextManager[Path] = importlib.resources.as_file(
            DATA_DIRECTORY.joinpath(str(cls.CHABRIER_DIRECTORY.joinpath(filename)))
        )
        with data as datapath:
            df: pd.DataFrame = pd.read_csv(
                datapath,  # type: ignore
                sep=r"\s+",
                comment="#",
                usecols=[0, 1, 2],  # type: ignore
                names=column_names,
                skiprows=2,
            )
        pivot_table: pd.DataFrame = df.pivot(index=T_name, columns=P_name, values=rho_name)
        log_T: Array = jnp.array(pivot_table.index.to_numpy())
        log_P: Array = jnp.array(pivot_table.columns.to_numpy())
        log_rho: Array = jnp.array(pivot_table.to_numpy())

        interpolator: RegularGridInterpolator = RegularGridInterpolator(
            (log_T, log_P), log_rho, method="linear"
        )

        def interpolator_hashable_function_wrapper(x) -> Array:
            """Converts interpolator to a hashable function"""
            return interpolator(x)

        return interpolator_hashable_function_wrapper

    @eqx.filter_jit
    def _convert_to_molar_density(self, log10_density_gcc: ArrayLike) -> Array:
        r"""Converts density to molar density

        Convert units: g/cm3 to mol/cm3 to mol/m3 for H2 (1e6 cm3 = 1 m3; 1 mol H2 = 2.016 g H2)

        Args:
            log10_density_gcc: Log10 density in g/cc

        Returns:
            Molar density in :math:`\mathrm{mol}\mathrm{m}^{-3}`
        """
        molar_density: Array = jnp.power(10, log10_density_gcc) / unit_conversion.cm3_to_m3
        composition_factor: float = (
            self.He_molar_mass_g_mol * self.He_fraction
            + self.H2_molar_mass_g_mol * (1 - self.He_fraction)
        )
        molar_density = molar_density / composition_factor

        return molar_density

    @override
    @eqx.filter_jit
    def log_fugacity(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """Log fugacity

        This performs a numerical integration to compute the fugacity, although an obvious
        speedup to implement is to precompute the fugacity (integral), either by calculating it
        during initialisation or by storing it in a lookup table that is read in.

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            Log fugacity in bar
        """
        temperature = jnp.asarray(temperature)
        log10_pressure: Array = jnp.log10(pressure)
        temperature, log10_pressure = jnp.broadcast_arrays(temperature, log10_pressure)

        # Pressure range to integrate over
        pressures: Array = jnp.logspace(
            jnp.log10(PRESSURE_REFERENCE), log10_pressure, num=self.integration_steps
        )
        # jax.debug.print("pressures.shape = {out}", out=pressures.shape)
        dP: Array = jnp.diff(pressures, axis=0)
        # jax.debug.print("dP.shape = {out}", out=dP.shape)

        volumes: Array = self.volume(temperature, pressures)
        # jax.debug.print("volumes.shape = {out}", out=volumes.shape)
        avg_volumes: Array = (volumes[:-1] + volumes[1:]) * 0.5
        # jax.debug.print("avg_volumes.shape = {out}", out=avg_volumes.shape)

        # Trapezoid integration
        volume_integral: Array = jnp.sum(avg_volumes * dP, axis=0)
        # jax.debug.print("volume_integral.shape = {out}", out=volume_integral.shape)

        log_fugacity: Array = volume_integral / (GAS_CONSTANT_BAR * temperature)

        return log_fugacity

    @override
    @eqx.filter_jit
    def volume(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        log10_density_gcc: Array = self.log10_density_func(
            (jnp.log10(temperature), jnp.log10(unit_conversion.bar_to_GPa * pressure))
        )
        # jax.debug.print("log10_density_gcc = {out}", out=log10_density_gcc)
        molar_density: Array = self._convert_to_molar_density(log10_density_gcc)
        volume: Array = jnp.reciprocal(molar_density)
        # jax.debug.print("volume = {out}", out=volume)

        return volume

    @override
    @eqx.filter_jit
    def volume_integral(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        log_fugacity: Array = self.log_fugacity(temperature, pressure)
        volume_integral: Array = log_fugacity * GAS_CONSTANT_BAR * temperature

        return volume_integral


calibration_chabrier21: ExperimentalCalibration = ExperimentalCalibration(
    temperature_min=100, temperature_max=1.0e8, pressure_min=None, pressure_max=1.0e17
)
"""Calibration for :cite:t:`CD21`"""
H2_chabrier21: RealGas = Chabrier.create(Path("TABLE_H_TP_v1"))
"""H2 :cite:p:`CD21`"""
H2_chabrier21_bounded: RealGas = CombinedRealGas.create(
    [H2_chabrier21],
    [calibration_chabrier21],
)
"""H2 bounded :cite:p:`CD21`"""
He_chabrier21: RealGas = Chabrier.create(Path("TABLE_HE_TP_v1"))
"""He :cite:p:`CD21`"""
He_chabrier21_bounded: RealGas = CombinedRealGas.create([He_chabrier21], [calibration_chabrier21])
"""He bounded :cite:p:`CD21`"""
H2_He_Y0275_chabrier21: RealGas = Chabrier.create(Path("TABLEEOS_2021_TP_Y0275_v1"))
"""H2HeY0275 :cite:p:`CD21`"""
H2_He_Y0275_chabrier21_bounded: RealGas = CombinedRealGas.create(
    [H2_He_Y0275_chabrier21], [calibration_chabrier21]
)
"""H2HeY0275 bounded :cite:p:`CD21`"""
H2_He_Y0292_chabrier21: RealGas = Chabrier.create(Path("TABLEEOS_2021_TP_Y0292_v1"))
"""H2HeY0292 :cite:p:`CD21`"""
H2_He_Y0292_chabrier21_bounded: RealGas = CombinedRealGas.create(
    [H2_He_Y0292_chabrier21], [calibration_chabrier21]
)
"""H2HeY0292 bounded :cite:p:`CD21`"""
H2_He_Y0297_chabrier21: RealGas = Chabrier.create(Path("TABLEEOS_2021_TP_Y0297_v1"))
"""H2HeY0297 :cite:p:`CD21`"""
H2_He_Y0297_chabrier21_bounded: RealGas = CombinedRealGas.create(
    [H2_He_Y0297_chabrier21], [calibration_chabrier21]
)
"""H2HeY0297 bounded :cite:p:`CD21`"""


def get_chabrier_eos_models() -> dict[str, RealGas]:
    """Gets a dictionary of EOS models

    Returns:
        Dictionary of EOS models
    """
    eos_models: dict[str, RealGas] = {}
    eos_models["H2_chabrier21"] = H2_chabrier21_bounded
    eos_models["H2_He_Y0275_chabrier21"] = H2_He_Y0275_chabrier21_bounded
    eos_models["H2_He_Y0292_chabrier21"] = H2_He_Y0292_chabrier21_bounded
    eos_models["H2_He_Y0297_chabrier21"] = H2_He_Y0297_chabrier21_bounded
    eos_models["He_chabrier21"] = He_chabrier21_bounded

    return eos_models



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Solubility laws for hydrogen species

For every law there should be a test in the test suite.
"""

import equinox as eqx
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from atmodeller.eos._chabrier import H2_chabrier21_bounded
from atmodeller.solubility.core import Solubility, SolubilityPowerLaw, SolubilityPowerLawLog10
from atmodeller.utilities import unit_conversion

try:
    from typing import override  # type: ignore valid for Python 3.12+
except ImportError:
    from typing_extensions import override  # Python 3.11 and earlier


H2_andesite_hirschmann12: Solubility = SolubilityPowerLawLog10(1.01058631, 0.60128868)
"""H2 in synthetic andesite :cite:p:`HWA12`

Log-scale linear fit to fH2 vs H2 concentration for andesite in :cite:t:`HWA12{Table 2}`.
Experiments conducted from 0.7-3 GPa at 1400 C.
"""

H2_basalt_hirschmann12: Solubility = SolubilityPowerLawLog10(1.10083602, 0.52413928)
"""H2 in synthetic basalt :cite:p:`HWA12`

Log-scale linear fit to fH2 vs. H2 concentration for basalt in :cite:t:`HWA12{Table 2}`.
Experiments conducted from 0.7-3 GPa, 1400 C.
"""

H2_silicic_melts_gaillard03: Solubility = SolubilityPowerLaw(0.163, 1.252)
"""Fe-H redox exchange in silicate glasses :cite:p:`GSM03`

Power law fit for fH2 vs. H2 (ppm-wt) from :cite:t:`GSM03{Table 4}` data. Experiments at
pressures from 0.02-70 bar, temperatures from 300-1000C.
"""

H2O_ano_dio_newcombe17: Solubility = SolubilityPowerLaw(727, 0.5)
"""H2O in anorthite-diopside-eutectic compositions :cite:p:`NBB17`

Power law from :cite:t:`NBB17{Figure 5(A)}` for anorthite-diopside glass. Experiments conducted
at 1 atm and 1350 C. Melts equilibrated in 1 atm furnace with H2/CO2 gas mixtures that spanned
fO2 from IW-3 to IW+4.8 and pH2/pH2O from 0.003-24.
"""

H2O_basalt_dixon95: Solubility = SolubilityPowerLaw(965, 0.5)
"""H2O in MORB liquids :cite:p:`DSH95`

Refitted data to a power law by Paolo Sossi (fitting :cite:t:`DSH95{Figure 4}`, TODO: CHECK).
Experiments conducted at 1200 C, 200-717 bars with pure H2O.
"""

H2O_basalt_mitchell17: Solubility = SolubilityPowerLaw(258.946, 0.669)
"""H2O in basaltic melt :cite:p:`MGO17`

Refitted the H2O wt. % vs. fH2O fitted line from :cite:t:`MGO17{Figure 8}` to a power-law.
Experiments conducted at 1200 C and 1000 MPa total pressure. This fit includes data from
their experiments and prior studies on H2O solubility in basaltic melt at 1200 C and P at or
below 600 MPa.
"""

H2O_lunar_glass_newcombe17: Solubility = SolubilityPowerLaw(683, 0.5)
"""H2O in lunar basalt :cite:p:`NBB17`

Power law from :cite:t:`NBB17{Figure 5(A)}` for Lunar glass. Experiments conducted at 1 atm and
1350 C. Melts equilibrated in 1-atm furnace with H2/CO2 gas mixtures that spanned fO2 from IW-3
to IW+4.8.
"""

H2O_peridotite_sossi23: Solubility = SolubilityPowerLaw(647, 0.5)
"""H2O in peridotite liquids :cite:p:`STB23`

Power law parameters in the abstract for peridotitic glasses. Experiments conducted at 2173 K
and 1 bar and range of fO2 from IW-1.9 to IW+6.0.
"""


class _H2_chachan18(Solubility):
    """H2 solubility :cite:p:`CS18`

    Args:
        f_calibration: Calibration fugacity
        T_calibration: Calibration temperature
        X_calibration: Mass fraction at calibration conditions
        T0: Arrhenius temperature factor in K, which expresses the repulsive interaction of the
            molecule with magma. Defaults to 4000 K, which is the middle of the range the authors
            explore (from 3000 K to 5000 K).

    Attributes:
        f_calibration: Calibration fugacity
        T_calibration: Calibration temperature
        X_calibration: Mass fraction at calibration conditions
        T0: Arrhenius temperature factor
        A: Constant
    """

    f_calibration: ArrayLike
    T_calibration: ArrayLike
    X_calibration: ArrayLike
    T0: float = 4000
    A: ArrayLike = eqx.field(init=False)

    def __post_init__(self):
        self.A = jnp.exp(
            (self.T0 / self.T_calibration) + jnp.log(self.X_calibration / self.f_calibration)
        )
        # jax.debug.print("A = ", self.A)

    @override
    @eqx.filter_jit
    def concentration(self, fugacity: ArrayLike, *, temperature: ArrayLike, **kwargs) -> Array:
        del kwargs
        mass_fraction: Array = jnp.exp(-self.T0 / temperature) * self.A * fugacity
        ppmw: Array = mass_fraction * unit_conversion.fraction_to_ppm

        return ppmw


H2_chachan18: Solubility = _H2_chachan18(
    f_calibration=1500, T_calibration=1673, X_calibration=0.001
)
"""H2 by combining theory and experiment :cite:p:`CS18`"""

# At 1 GPa in the presence of pure H2, the molecular H2 concentration is 0.19 wt.%"
# Need to convert pressure to H2 fugacity
# With Chabrier EOS, f_calibration = 23986.034649111516
# With Zhang and Duan EOS, f_calibration2 = 28421.194323648964
T_calibration: float = 1673
P_calibration: float = 1 * unit_conversion.GPa_to_bar
f_calibration: ArrayLike = H2_chabrier21_bounded.fugacity(T_calibration, P_calibration)
X_calibration: float = 0.0019

H2_kite19: Solubility = _H2_chachan18(
    f_calibration=f_calibration, T_calibration=T_calibration, X_calibration=X_calibration
)
"""H2 by combining theory and experiment :cite:p:`KFS19`."""



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Solubility laws for other species

For every law there should be a test in the test suite.
"""

import equinox as eqx
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from atmodeller.constants import GAS_CONSTANT_BAR
from atmodeller.interfaces import RedoxBufferProtocol
from atmodeller.solubility.core import (
    Solubility,
    SolubilityPowerLaw,
    fO2_temperature_correction,
    power_law,
)
from atmodeller.thermodata import IronWustiteBuffer
from atmodeller.utilities import safe_exp, unit_conversion

try:
    from typing import override  # type: ignore valid for Python 3.12+
except ImportError:
    from typing_extensions import override  # Python 3.11 and earlier


Cl2_ano_dio_for_thomas21: Solubility = SolubilityPowerLaw(
    140.52 * unit_conversion.percent_to_ppm, 0.5
)
"""Cl in silicate melts :cite:p:`TW21`

Solubility law from :cite:t:`TW21{Figure 4}` showing relation between dissolved Cl concentration
and Cl fugacity for CMAS composition (An50Di28Fo22 (anorthite-diopside-forsterite), Fe-free 
low-degree mantle melt) at 1400 C and 1.5 GPa. Experiments from 0.5-2 GPa and 1200-1500 C.
"""

Cl2_basalt_thomas21: Solubility = SolubilityPowerLaw(78.56 * unit_conversion.percent_to_ppm, 0.5)
"""Cl in silicate melts :cite:p:`TW21`

Solubility law from :cite:t:`TW21{Figure 4}` showing relation between dissolved Cl concentration
and Cl fugacity for Icelandic basalt at 1400 C and 1.5 GPa. Experiments from 0.5-2 GPa and 
1200-1500 C.
"""

_He_henry_sol_constant_jambon86: float = 56e-5  # cm3*STP/g*bar
# Convert Henry solubility constant to mol/g*bar, 2.24e4 cm^3/mol at STP
# Convert He conc from mol/g to g He/g total and then to ppmw
_He_henry_sol_constant_jambon86 = (
    (_He_henry_sol_constant_jambon86 / 2.24e4) * 4.0026 * unit_conversion.fraction_to_ppm
)
He_basalt_jambon86: Solubility = SolubilityPowerLaw(_He_henry_sol_constant_jambon86, 1)
"""Solubility of He in tholeittic basalt melt :cite:p:`JWB86`

Experiments determined Henry's law solubility constant in tholetiitic basalt melt at 1 bar and
1250-1600 C. Using Henry's Law solubility constant for He from the abstract, convert from STP
units to mol/g*bar.
"""

_Ar_henry_sol_constant_jambon86: float = 5.9e-5  # cm3*STP/g*bar
# Convert Henry solubility constant to mol/g*bar, 2.24e4 cm^3/mol at STP
# Convert Ar conc from mol/g to g Ar/g total and then to ppmw
_Ar_henry_sol_constant_jambon86 = (
    (_Ar_henry_sol_constant_jambon86 / 2.24e4) * 39.948 * unit_conversion.fraction_to_ppm
)
Ar_basalt_jambon86: Solubility = SolubilityPowerLaw(_Ar_henry_sol_constant_jambon86, 1)
"""Solubility of Ar in tholeittic basalt melt :cite:p:`JWB86`

Experiments determined Henry's law solubility constant in tholetiitic basalt melt at 1 bar and
1250-1600 C. Using Henry's Law solubility constant for Ar from the abstract, convert from STP
units to mol/g*bar.
"""

_Ne_henry_sol_constant_jambon86: float = 25e-5  # cm3*STP/g*bar
# Convert Henry solubility constant to mol/g*bar, 2.24e4 cm^3/mol at STP
# Convert Ne conc from mol/g to g Ne/g total and then to ppmw
_Ne_henry_sol_constant_jambon86 = (
    (_Ne_henry_sol_constant_jambon86 / 2.24e4) * 20.1797 * unit_conversion.fraction_to_ppm
)
Ne_basalt_jambon86: Solubility = SolubilityPowerLaw(_Ne_henry_sol_constant_jambon86, 1)
"""Solubility of Ne in tholeittic basalt melt :cite:p:`JWB86`

Experiments determined Henry's law solubility constant in tholetiitic basalt melt at 1 bar and
1250-1600 C. Using Henry's Law solubility constant for Ne from the abstract, convert from STP
units to mol/g*bar.
"""

_Kr_henry_sol_constant_jambon86: float = 3.0e-5  # cm3*STP/g*bar
# Convert Henry solubility constant to mol/g*bar, 2.24e4 cm^3/mol at STP
# Convert Kr conc from mol/g to g Kr/g total and then to ppmw
_Kr_henry_sol_constant_jambon86 = (
    (_Kr_henry_sol_constant_jambon86 / 2.24e4) * 83.798 * unit_conversion.fraction_to_ppm
)
Kr_basalt_jambon86: Solubility = SolubilityPowerLaw(_Kr_henry_sol_constant_jambon86, 1)
"""Solubility of Kr in tholeittic basalt melt :cite:p:`JWB86`

Experiments determined Henry's law solubility constant in tholetiitic basalt melt at 1 bar and
1250-1600 C. Using Henry's Law solubility constant for Kr from the abstract, convert from STP
units to mol/g*bar.
"""

_Xe_henry_sol_constant_jambon86: float = 1.7e-5  # cm3*STP/g*bar
# Convert Henry solubility constant to mol/g*bar, 2.24e4 cm^3/mol at STP
# Convert Xe conc from mol/g to g Xe/g total and then to ppmw
_Xe_henry_sol_constant_jambon86 = (
    (_Xe_henry_sol_constant_jambon86 / 2.24e4) * 131.293 * unit_conversion.fraction_to_ppm
)
Xe_basalt_jambon86: Solubility = SolubilityPowerLaw(_Xe_henry_sol_constant_jambon86, 1)
"""Solubility of Xe in tholeittic basalt melt :cite:p:`JWB86`

Experiments determined Henry's law solubility constant in tholetiitic basalt melt at 1 bar and
1250-1600 C. Using Henry's Law solubility constant for Xe from the abstract, convert from STP
units to mol/g*bar.
"""


class _N2_basalt_bernadou21(Solubility):
    """N2 in basaltic silicate melt :cite:p:`BGF21`

    :cite:t:`BGF21{Equation 18}` and using :cite:t:`BGF21{Equations 19-20}` and the values for the
    thermodynamic constants from :cite:t:`BGF21{Table 6}`. Experiments on basaltic samples at fluid
    saturation in C-H-O-N system, pressure range: 0.8-10 kbar, temperature range: 1200-1300 C;
    fO2 range: IW+4.9 to IW-4.7. Using their experimental results and a database for N
    concentrations at fluid saturation from 1 bar to 10 kbar, calibrated their solubility law.
    """

    @override
    @eqx.filter_jit
    def concentration(
        self,
        fugacity: ArrayLike,
        *,
        temperature: ArrayLike,
        pressure: ArrayLike,
        fO2: ArrayLike,
    ) -> Array:
        # Numerator and denominator of k13 and k14 should both have units of J/mol so that k13 and
        # k14 are unitless
        k13: Array = safe_exp(
            -(29344 + 121 * temperature + 4 * pressure) / (GAS_CONSTANT_BAR * 1.0e5 * temperature)
        )
        k14: Array = safe_exp(
            -(183733 + 172 * temperature - 5 * pressure) / (GAS_CONSTANT_BAR * 1.0e5 * temperature)
        )
        molfrac: Array = (k13 * fugacity) + ((fO2 ** (-3 / 4)) * k14 * (fugacity**0.5))
        ppmw: Array = molfrac * unit_conversion.fraction_to_ppm

        return ppmw


N2_basalt_bernadou21: Solubility = _N2_basalt_bernadou21()


class _N2_basalt_dasgupta22(Solubility):
    """N2 in silicate melts :cite:p:`DFP22`

    Using :cite:t:`DFP22{Equation 10}`, composition parameters from :cite:t:`DFP22{Figure 8}`, and
    Iron-wustite buffer (logIW_fugacity) from :cite:t:`OP93,HGD08`.

    Performed experiments on 80:20 synthetic basalt-Si3N4 mixture at 1.5-3.0 GPa and 1300-1600 C
    fO2 from ~IW-3 to IW-4. Combined this high pressure data with lower pressure studies to derive
    their N solubility law.

    Args:
        xsio2: Mole fraction of SiO2. Defaults to 0.56.
        xal2o3: Mole fraction of Al2O3. Defaults to 0.11.
        xtio2: Mole fraction of TiO2. Defaults to 0.01.

    Attributes:
        xsio2: Mole fraction of SiO2
        xal2o3: Mole fraction of Al2O3
        xtio2: Mole fraction of TiO2
    """

    xsio2: float = 0.56
    xal2o3: float = 0.11
    xtio2: float = 0.01
    # The buffer is evaluated at the pressure of interest, not the default of 1 bar
    _buffer: RedoxBufferProtocol = eqx.field(init=False)

    def __post_init__(self):
        self._buffer = IronWustiteBuffer(evaluation_pressure=None)

    @override
    @eqx.filter_jit
    def concentration(
        self,
        fugacity: ArrayLike,
        *,
        temperature: ArrayLike,
        pressure: ArrayLike,
        fO2: ArrayLike,
    ) -> Array:
        fugacity_gpa: ArrayLike = fugacity * unit_conversion.bar_to_GPa
        pressure_gpa: ArrayLike = pressure * unit_conversion.bar_to_GPa

        fo2_shift: Array = jnp.log10(fO2) - self._buffer.log10_fugacity_buffer(
            temperature, pressure
        )
        ppmw: Array = safe_exp(
            (5908.0 * jnp.sqrt(pressure_gpa) / temperature) - (1.6 * fo2_shift)
        ) * jnp.sqrt(fugacity_gpa)
        ppmw = ppmw + fugacity_gpa * safe_exp(
            4.67 + (7.11 * self.xsio2) - (13.06 * self.xal2o3) - (120.67 * self.xtio2)
        )

        return ppmw


N2_basalt_dasgupta22: Solubility = _N2_basalt_dasgupta22()
"""N2 in basaltic silicate melt :cite:p:`BGF21`

:cite:t:`BGF21{Equation 18}` and using :cite:t:`BGF21{Equations 19-20}` and the values for the
thermodynamic constants from :cite:t:`BGF21{Table 6}`. Experiments on basaltic samples at fluid
saturation in C-H-O-N system, pressure range: 0.8-10 kbar, temperature range: 1200-1300 C;
fO2 range: IW+4.9 to IW-4.7. Using their experimental results and a database for N
concentrations at fluid saturation from 1 bar to 10 kbar, calibrated their solubility law.
"""


class _N2_basalt_libourel03(Solubility):
    """N2 in basalt (tholeiitic) magmas :cite:p:`LMH03`

    :cite:t:`LMH03{Equation 23}`, includes dependencies on fN2 and fO2. Experiments conducted at 1
    atm and 1425 C (two experiments at 1400 C), fO2 from IW-8.3 to IW+8.7 using mixtures of CO, CO2
    and N2 gases.
    """

    @override
    @eqx.filter_jit
    def concentration(
        self,
        fugacity: ArrayLike,
        *,
        temperature: ArrayLike,
        pressure: ArrayLike,
        fO2: ArrayLike,
    ) -> Array:
        # Libourel performed the experiment at 1698 K and fitted for fO2 at this temperature hence
        # a correction is necessary.
        libourel_temperature: ArrayLike = 1698.15
        adjusted_fO2: Array = fO2_temperature_correction(
            fO2,
            temperature=temperature,
            pressure=pressure,
            reference_temperature=libourel_temperature,
        )

        ppmw: Array = power_law(fugacity, 0.0611, 1)
        constant: Array = power_law(adjusted_fO2, 5.97e-10, -0.75)
        ppmw2: Array = power_law(fugacity, constant, 0.5)
        ppmw = ppmw + ppmw2

        return ppmw


N2_basalt_libourel03: Solubility = _N2_basalt_libourel03()
"""N2 in basalt (tholeiitic) magmas :cite:p:`LMH03`

:cite:t:`LMH03{Equation 23}`, includes dependencies on fN2 and fO2. Experiments conducted at 1
atm and 1425 C (two experiments at 1400 C), fO2 from IW-8.3 to IW+8.7 using mixtures of CO, CO2
and N2 gases.
"""



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Real gas EOS from :cite:t:`HP91,HP98,HP11`"""

from __future__ import annotations

import logging
from abc import abstractmethod
from typing import Callable

import equinox as eqx
import jax.numpy as jnp
from jax import Array, lax
from jax.typing import ArrayLike
from scipy.constants import kilo

from atmodeller.constants import GAS_CONSTANT_BAR
from atmodeller.eos._aggregators import CombinedRealGas
from atmodeller.eos.core import (
    CORK,
    RealGas,
    RedlichKwongABC,
    RedlichKwongImplicitDenseFluidABC,
    RedlichKwongImplicitGasABC,
    VirialCompensation,
)
from atmodeller.thermodata import CriticalData, select_critical_data
from atmodeller.utilities import ExperimentalCalibration

try:
    from typing import override  # type: ignore valid for Python 3.12+
except ImportError:
    from typing_extensions import override  # Python 3.11 and earlier


logger: logging.Logger = logging.getLogger(__name__)


class CorrespondingStatesUnitConverter:
    r"""Unit converter for Holland and Powell CORK corresponding states model

    This converts the coefficient units from Holland and Powell to the required units for
    Atmodeller. This accounts for kilo factors and converting the energy from J to SI volume and
    pressure in bar using:

    .. math::
        1\ \mathrm{J} = 10^{-5}\ \mathrm{m}^3\ \mathrm{bar}
    """

    @staticmethod
    def convert_a_coefficients(a_coefficients: tuple[float, ...]) -> tuple[float, ...]:
        r"""Converts the a coefficients for corresponding states.

        The a coefficients (a0 and a1) have units :cite:p:`HP91{Equation 9}`:

        .. math::
            \left(\frac{\mathrm{kJ}}{\mathrm{mol}\ \mathrm{K}}\right)^2

        Converting kJ gives rise to a unit conversion factor of :math:`10^{-4}`.

        Args:
            a_coefficients: a coefficients from Holland and Powell

        Returns:
            a coefficients with units of

            .. math::
                \left(\frac{\mathrm{m}^3\ \mathrm{bar}}{\mathrm{mol}\ \mathrm{K}}\right)^2
        """
        factor: float = 1e-4

        return tuple(map(lambda a_coefficient: factor * a_coefficient, a_coefficients))

    @staticmethod
    def convert_b_coefficient(b_coefficient: float) -> float:
        r"""Converts the b coefficient for corresponding states.

        The b coefficient (b0) has units :cite:p:`HP91{Equation 9}`:

        .. math::

            \frac{\mathrm{kJ}}{\mathrm{mol}\ \mathrm{K}}

        Converting kJ gives rise to a unit conversion factor of :math:`10^{-2}`.

        Args:
            b_coefficient: b coefficient from Holland and Powell

        Returns:
            b coefficient with units of

            .. math::
                \frac{\mathrm{m}^3\ \mathrm{bar}}{\mathrm{mol}\ \mathrm{K}}
        """
        factor: float = 1e-2

        return b_coefficient * factor

    @staticmethod
    def convert_virial_coefficients(virial_coefficients: tuple[float, ...]) -> tuple[float, ...]:
        r"""Converts the virial coefficients for corresponding states

        The virial coefficients, for example associated with coefficients c and d in
        :cite:`HP91{Table 2}`, have units:

        .. math::

            \frac{\mathrm{kJ}}{\mathrm{mol}\ \mathrm{K}}

        Converting kJ gives rise to a unit conversion factor of :math:`10^{-2}`.

        Args:
            virial_coefficient: Virial coefficients from Holland and Powell

        Returns:
            virial coefficients with units of

            .. math::
                \frac{\mathrm{m}^3\ \mathrm{bar}}{\mathrm{mol}\ \mathrm{K}}
        """
        factor: float = 1e-2

        return tuple(
            map(lambda virial_coefficient: factor * virial_coefficient, virial_coefficients)
        )


class FullUnitConverter:
    r"""Unit converter for Holland and Powell full CORK models for H2O and CO2

    This converts the coefficient units from Holland and Powell to the required units for
    Atmodeller. This accounts for kilo factors and converting the energy from J to SI volume and
    pressure in bar using:

    .. math::
        1\ \mathrm{J} = 10^{-5}\ \mathrm{m}^3\ \mathrm{bar}
    """

    @staticmethod
    def convert_a_coefficients(a_coefficients: tuple[float, ...]) -> tuple[float, ...]:
        r"""Converts the a coefficients for the full CORK models

        The a parameter has units :cite:p:`HP91{Table 1}`

        .. math::
            \frac{\mathrm{kJ}^2\ \mathrm{K}^{1/2}}{\mathrm{kbar}\ \mathrm{mol}^2}

        Each a coefficient has a different power of K, but this is dealt with during the
        multiplication with temperature. Here, we just need to deal with kJ and other kilo factors,
        which gives rise to a unit conversion factor of :math:`10^{-7}`.

        Args:
            a_coefficients: a coefficients from Holland and Powell

        Returns:
            a coefficients with units converted such that the a parameter has units of

            .. math::
                \left(\frac{\mathrm{m}^3}{\mathrm{mol}}\right)^2\ \mathrm{bar}\ \mathrm{K}^{1/2}
        """
        factor: float = 1e-7

        return tuple(map(lambda a_coefficient: factor * a_coefficient, a_coefficients))

    @staticmethod
    def convert_b_coefficient(b_coefficient: float) -> float:
        r"""Converts the b coefficient for the full CORK models

        The b parameter has units :cite:p:`HP91{Table 1}`

        .. math::
            \frac{\mathrm{kJ}}{\mathrm{kbar}\ \mathrm{mol}}

        Converting J gives rise to a unit conversion factor of :math:`10^{-5}`.

        Args:
            b_coefficient: b coefficient from Holland and Powell

        Returns:
            b coefficients with units converted such that the b parameter has units of

            .. math::
                \frac{\mathrm{m}^3}{\mathrm{mol}}
        """
        factor: float = 1e-5

        return b_coefficient * factor

    @staticmethod
    def convert_virial_coefficients(
        virial_coefficients: tuple[float, ...], pressure_exponent
    ) -> tuple[float, ...]:
        r"""Converts the virial coefficients for the full CORK models

        The volume correction to the MRK volume, :math:`\Delta V`, based on the parameter that the
        coefficients determine is given by:

        .. math::
            \Delta V = \mathrm{parameter(coefficients)} (P-P_0)^\gamma

        where :math:`\gamma` is `pressure_exponent` and :math:`P_0` is the pressure at which the
        MRK equation begins to overestimate the molar volume significantly.

        Args:
            virial_coefficients: Virial coefficients from Holland and Powell
            pressure_exponent: Pressure exponent :math:`\gamma`, also given by Holland and Powell

        Returns:
            Virial coefficients with the appropriate units for Atmodeller
        """
        factor: float = 1e-5 / kilo**pressure_exponent

        return tuple(
            map(lambda virial_coefficient: factor * virial_coefficient, virial_coefficients)
        )


class MRKCorrespondingStatesHP91(RedlichKwongABC):
    """MRK corresponding states :cite:p:`HP91`

    Universal constants from :cite:t:`HP91{Table 2}`

    Args:
        critical_data: Critical data
    """

    critical_data: CriticalData
    _a_coefficients: tuple[float, ...] = eqx.field(init=False)
    _b: float = eqx.field(init=False)

    def __post_init__(self):
        self._a_coefficients = CorrespondingStatesUnitConverter.convert_a_coefficients(
            (5.45963e-5, -8.63920e-6, 0)
        )
        self._b = CorrespondingStatesUnitConverter.convert_b_coefficient(9.18301e-4)

    @property
    def critical_pressure(self) -> float:
        """Critical pressure in bar"""
        return self.critical_data.pressure

    @property
    def critical_temperature(self) -> float:
        """Critical temperature in K"""
        return self.critical_data.temperature

    @classmethod
    def get_species(cls, species: str) -> MRKCorrespondingStatesHP91:
        """Gets an MRK corresponding states model for a given species.

        Args:
            species: Species

        Returns:
            An MRK corresponding states model for the species
        """
        critical_data: CriticalData = select_critical_data(species)

        return cls(critical_data)

    @override
    @eqx.filter_jit
    def a(self, temperature: ArrayLike, pressure: ArrayLike) -> ArrayLike:
        r"""MRK `a` parameter :cite:p:`HP91{Equation 9}`

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            MRK `a` parameter in
            :math:`(\mathrm{m}^3\ \mathrm{mol}^{-1})^2\ \mathrm{K}^{1/2}\ \mathrm{bar}`
        """
        del pressure

        a: ArrayLike = (
            self._a_coefficients[0] * jnp.power(self.critical_temperature, (5.0 / 2))
            + self._a_coefficients[1]
            * jnp.power(self.critical_temperature, (3.0 / 2))
            * temperature
            + self._a_coefficients[2]
            * jnp.power(self.critical_temperature, (1.0 / 2))
            * jnp.square(temperature)
        )
        a = a / self.critical_pressure

        return a

    @override
    @eqx.filter_jit
    def b(self) -> ArrayLike:
        r"""MRK `b` parameter computed from :attr:`b0` :cite:p:`HP91{Equation 9}`.

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            MRK `b` parameter in :math:`\mathrm{m}^3\ \mathrm{mol}^{-1}`.
        """
        b: ArrayLike = self._b * self.critical_temperature / self.critical_pressure

        return b


class MRKImplicitHP91ABCMixin(eqx.Module):
    """MRK implicit :cite:p:`HP91`

    Universal constants from :cite:t:`HP91{Table 1}`.

    Args:
        a_coefficients: `a` coefficients
        b: `b` coefficient
        Ta: Temperature at which the `a` parameter is equal for the dense fluid and gas in K
        Tc: Critical temperature in K
    """

    _a_coefficients: tuple[float, ...]
    _b: float
    _Ta: float
    _Tc: float

    @abstractmethod
    def delta_temperature_for_a(self, temperature: ArrayLike) -> ArrayLike:
        """Temperature difference for the calculation of the `a` parameter

        Args:
            temperature: Temperature in K

        Returns:
            Temperature difference in K
        """
        ...

    @eqx.filter_jit
    def a(self, temperature: ArrayLike, pressure: ArrayLike) -> ArrayLike:
        r"""MRK `a` parameter :cite:p:`HP91{Equation 6}`

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            MRK `a` parameter
        """
        del pressure

        delta_temperature: ArrayLike = self.delta_temperature_for_a(temperature)
        a: ArrayLike = (
            self._a_coefficients[0]
            + self._a_coefficients[1] * delta_temperature
            + self._a_coefficients[2] * jnp.square(delta_temperature)
            + self._a_coefficients[3] * jnp.power(delta_temperature, 3)
        )

        return a

    @eqx.filter_jit
    def b(self) -> ArrayLike:
        return self._b


class MRKImplicitGasHP91(MRKImplicitHP91ABCMixin, RedlichKwongImplicitGasABC):
    """MRK for gaseous phase :cite:p:`HP91{Equation 6a}`"""

    @override
    def delta_temperature_for_a(self, temperature: ArrayLike) -> ArrayLike:
        return self._Ta - temperature


Tc_H2O: float = 695
"""Critical temperature of H2O in K for the MRK/CORK model :cite:p:`HP91`"""
Ta_H2O: float = 673  # K
r"""Temperature at which :math:`a_{\mathrm gas} = a` for H2O by fitting :cite:p:`HP91`"""
b0_H2O: float = FullUnitConverter.convert_b_coefficient(1.465)
"""b parameter value which is the same across all H2O phases :cite:p:`HP91`"""

H2OMrkGasHolland91: MRKImplicitGasHP91 = MRKImplicitGasHP91(
    FullUnitConverter.convert_a_coefficients((1113.4, 5.8487, -2.1370e-2, 6.8133e-5)),
    b0_H2O,
    Ta_H2O,
    Tc_H2O,
)
"""H2O MRK for gas phase :cite:p:`HP91`"""


class MRKImplicitLiquidHP91(MRKImplicitHP91ABCMixin, RedlichKwongImplicitDenseFluidABC):
    """MRK for liquid phase :cite:p`HP91{Equation 6}`"""

    @override
    @eqx.filter_jit
    def delta_temperature_for_a(self, temperature: ArrayLike) -> ArrayLike:
        return self._Ta - temperature


H2OMrkLiquidHolland91: MRKImplicitLiquidHP91 = MRKImplicitLiquidHP91(
    FullUnitConverter.convert_a_coefficients((1113.4, -0.88517, 4.53e-3, -1.3183e-5)),
    b0_H2O,
    Ta_H2O,
    Tc_H2O,
)
"""H2O MRK for liquid phase :cite:p`HP91`"""


class MRKImplicitFluidHP91(MRKImplicitHP91ABCMixin, RedlichKwongImplicitDenseFluidABC):
    """MRK for supercritical fluid :cite:p:`HP91{Equation 6}`"""

    @override
    @eqx.filter_jit
    def delta_temperature_for_a(self, temperature: ArrayLike) -> ArrayLike:
        return temperature - self._Ta

    @override
    @eqx.filter_jit
    def initial_volume(self, temperature: ArrayLike, pressure: ArrayLike) -> ArrayLike:
        r"""Initial guess volume to ensure convergence to the correct root

        See :cite:t:`HP91{Appendix}`. It appears that there is only ever a single root, even if
        Ta < temperature < Tc. Holland and Powell state that a single root exists if
        temperature > Tc, but this appears to be true if temperature > Ta. Nevertheless, the
        initial guess is changed accordingly.

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            Initial volume in :math:`\mathrm{m}^3\ \mathrm{mol}^{-1}`
        """

        def low_temperature_case() -> ArrayLike:
            return self.b() / 2

        def high_temperature_case() -> ArrayLike:
            return GAS_CONSTANT_BAR * temperature / pressure + self.b()

        initial_volume: Array = jnp.where(
            temperature < jnp.array(self._Tc),
            low_temperature_case(),
            high_temperature_case(),
        )

        return initial_volume


H2OMrkFluidHolland91: MRKImplicitFluidHP91 = MRKImplicitFluidHP91(
    FullUnitConverter.convert_a_coefficients(
        (
            1113.4,
            -0.22291,
            -3.8022e-4,
            1.7791e-7,
        )
    ),
    b0_H2O,
    Ta_H2O,
    Tc_H2O,
)
"""H2O MRK for fluid phase :cite:p`HP91`"""

CO2_critical_data: CriticalData = select_critical_data("CO2_g")
"""Alternative values from :cite:t:`HP91` are 304.2 K and 73.8 bar"""
CO2MrkHolland91: MRKImplicitFluidHP91 = MRKImplicitFluidHP91(
    FullUnitConverter.convert_a_coefficients((741.2, -0.10891, -3.4203e-4, 0)),
    FullUnitConverter.convert_b_coefficient(3.057),
    0,
    CO2_critical_data.temperature,
)
"""CO2 MRK :cite:p:`HP91{Above Equation 7}`

Critical behaviour is not considered for CO2 by :cite:t:`HP91`, but for consistency with the 
formulation for H2O, the CO2 critical temperature is set.
"""


class H2OMrkGasFluid91(RealGas):
    """A MRK model for H2O for the gas and supercritical fluid

    Args:
        mrk_fluid: The MRK for the supercritical fluid
        mrk_gas: The MRK for the subcritical gas
        Ta: Temperature at which a_gas = a in the MRK formulation in K
        Tc: Critical temperature in K
    """

    mrk_fluid: MRKImplicitFluidHP91 = H2OMrkFluidHolland91
    """The MRK for the supercritical fluid"""
    mrk_gas: MRKImplicitGasHP91 = H2OMrkGasHolland91
    """The MRK for the subcritical gas"""
    Ta: float = Ta_H2O
    """ Temperature at which a_gas = a in the MRK formulation in K"""
    Tc: float = Tc_H2O
    """Critical temperature in K"""

    @eqx.filter_jit
    def _select_condition(self, temperature: ArrayLike) -> Array:
        """Selects the condition

        Args:
            temperature: Temperature in K

        Returns:
            Integer denoting the condition, i.e. the region of phase space
        """
        temperature_array: Array = jnp.asarray(temperature)

        # Supercritical
        cond0: Array = temperature_array >= self.Tc
        # jax.debug.print("cond0 = {cond}", cond=cond0)
        # Below Ta
        cond1: Array = temperature_array <= self.Ta
        # jax.debug.print("cond1 = {cond}", cond=cond1)
        # Below Tc
        cond2: Array = temperature_array < self.Tc
        # Ensure cond2 is exclusive of cond1
        cond2 = jnp.logical_and(cond2, ~cond1)

        # All conditions are mutually exclusive
        condition: Array = jnp.select([cond0, cond1, cond2], [0, 1, 2])
        # jax.debug.print("condition = {condition}", condition=condition)

        return condition

    @override
    @eqx.filter_jit
    def volume_integral(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        r"""Volume integral :cite:p:`HP91{Appendix A}`

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            Volume integral in :math:`\mathrm{m}^3\ \mathrm{bar}\ \mathrm{mol}^{-1}`
        """
        condition: Array = self._select_condition(temperature)

        def volume_integral0() -> Array:
            return self.mrk_fluid.volume_integral(temperature, pressure)

        def volume_integral1() -> Array:
            return self.mrk_gas.volume_integral(temperature, pressure)

        def volume_integral2() -> Array:
            return self.mrk_fluid.volume_integral(temperature, pressure)

        volume_integral_funcs: list[Callable] = [
            volume_integral0,
            volume_integral1,
            volume_integral2,
        ]

        volume_integral: Array = lax.switch(condition, volume_integral_funcs)
        # jax.debug.print("volume_integral = {out}", out=volume_integral)

        return volume_integral

    @override
    @eqx.filter_jit
    def log_fugacity(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        r"""Log fugacity :cite:p:`HP91{Equation 8}`

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            Log fugacity
        """
        log_fugacity: Array = self.volume_integral(temperature, pressure) / (
            GAS_CONSTANT_BAR * temperature
        )

        return log_fugacity

    @override
    @eqx.filter_jit
    def volume(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        r"""Volume

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            Volume in :math:`\mathrm{m}^3\ \mathrm{mol}^{-1}`
        """
        condition: Array = self._select_condition(temperature)

        def volume0() -> ArrayLike:
            return self.mrk_fluid.volume(temperature, pressure)

        def volume1() -> ArrayLike:
            return self.mrk_gas.volume(temperature, pressure)

        def volume2() -> ArrayLike:
            return self.mrk_fluid.volume(temperature, pressure)

        volume_funcs: list[Callable] = [volume0, volume1, volume2]

        volume: Array = lax.switch(condition, volume_funcs)
        # jax.debug.print("volume = {out}", out=volume)

        return volume


class H2OMrkHP91(RealGas):
    """A MRK model for H2O that accommodates critical behaviour

    Args:
        mrk_fluid: The MRK for the supercritical fluid
        mrk_gas: The MRK for the subcritical gas
        mrk_liquid: The MRK for the subcritical liquid
        Ta: Temperature at which a_gas = a in the MRK formulation in K
        Tc: Critical temperature in K
    """

    mrk_fluid: MRKImplicitFluidHP91 = H2OMrkFluidHolland91
    """The MRK for the supercritical fluid"""
    mrk_gas: MRKImplicitGasHP91 = H2OMrkGasHolland91
    """The MRK for the subcritical gas"""
    mrk_liquid: MRKImplicitLiquidHP91 = H2OMrkLiquidHolland91
    """The MRK for the subcritical liquid"""
    Ta: float = Ta_H2O
    """Temperature at which a_gas = a in the MRK formulation in K"""
    Tc: float = Tc_H2O
    """Critical temperature in K"""

    @eqx.filter_jit
    def Psat(self, temperature: ArrayLike) -> Array:
        """Saturation curve

        Compared to :cite:t:`HP91` the pressure is returned in bar, as required by Atmodeller.

        Args:
            temperature: Temperature in K

        Returns:
            Saturation curve pressure in bar
        """
        Psat: Array = (
            -13.627
            + 7.29395e-4 * jnp.square(temperature)
            - 2.34622e-6 * jnp.power(temperature, 3)
            + 4.83607e-12 * jnp.power(temperature, 5)
        )

        return Psat

    @eqx.filter_jit
    def _select_condition(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """Selects the condition

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            Integer denoting the condition, i.e. the region of phase space
        """
        Psat: Array = self.Psat(temperature)
        temperature_array: Array = jnp.asarray(temperature)
        pressure_array: Array = jnp.asarray(pressure)

        # Supercritical (saturation pressure irrelevant)
        cond0: Array = temperature_array >= self.Tc
        # jax.debug.print("cond0 = {cond}", cond=cond0)
        # Below the saturation pressure and below Ta
        cond1: Array = jnp.logical_and(temperature_array <= self.Ta, pressure_array <= Psat)
        # jax.debug.print("cond1 = {cond}", cond=cond1)
        # Below the saturation pressure and below Tc
        cond2: Array = jnp.logical_and(temperature_array < self.Tc, pressure_array <= Psat)
        # Ensure cond2 is exclusive of cond1
        cond2 = jnp.logical_and(cond2, ~cond1)
        # jax.debug.print("cond2 = {cond}", cond=cond2)
        # Above the saturation pressure and below Ta
        cond3: Array = jnp.logical_and(temperature_array <= self.Ta, pressure_array > Psat)
        # jax.debug.print("cond3 = {cond}", cond=cond3)
        # Above the saturation pressure and below Tc
        cond4: Array = jnp.logical_and(temperature_array < self.Tc, pressure_array > Psat)
        # Ensure cond4 is exclusive of cond3
        cond4 = jnp.logical_and(cond4, ~cond3)
        # jax.debug.print("cond4 = {cond}", cond=cond4)

        # All conditions are mutually exclusive
        condition: Array = jnp.select([cond0, cond1, cond2, cond3, cond4], [0, 1, 2, 3, 4])
        # jax.debug.print("condition = {condition}", condition=condition)

        return condition

    @override
    @eqx.filter_jit
    def volume_integral(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        r"""Volume integral :cite:p:`HP91{Appendix A}`

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            Volume integral in :math:`\mathrm{m}^3\ \mathrm{bar}\ \mathrm{mol}^{-1}`
        """
        condition: Array = self._select_condition(temperature, pressure)
        Psat: Array = self.Psat(temperature)

        def volume_integral0() -> Array:
            return self.mrk_fluid.volume_integral(temperature, pressure)

        def volume_integral1() -> Array:
            return self.mrk_gas.volume_integral(temperature, pressure)

        def volume_integral2() -> Array:
            return self.mrk_fluid.volume_integral(temperature, pressure)

        def volume_integral3() -> Array:
            value: Array = self.mrk_gas.volume_integral(temperature, Psat)
            value = value - self.mrk_liquid.volume_integral(temperature, Psat)
            value = value + self.mrk_liquid.volume_integral(temperature, pressure)

            return value

        def volume_integral4() -> Array:
            return self.mrk_fluid.volume_integral(temperature, pressure)

        volume_integral_funcs: list[Callable] = [
            volume_integral0,
            volume_integral1,
            volume_integral2,
            volume_integral3,
            volume_integral4,
        ]

        volume_integral: Array = lax.switch(condition, volume_integral_funcs)
        # jax.debug.print("volume_integral = {out}", out=volume_integral)

        return volume_integral

    @override
    @eqx.filter_jit
    def log_fugacity(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        r"""Log fugacity :cite:p:`HP91{Equation 8}`

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            Log fugacity
        """
        log_fugacity: Array = self.volume_integral(temperature, pressure) / (
            GAS_CONSTANT_BAR * temperature
        )

        return log_fugacity

    @override
    @eqx.filter_jit
    def volume(self, temperature: ArrayLike, pressure: ArrayLike) -> ArrayLike:
        r"""Volume

        Args:
            temperature: Temperature in K
            pressure: Pressure in bar

        Returns:
            Volume in :math:`\mathrm{m}^3\ \mathrm{mol}^{-1}`
        """
        condition: Array = self._select_condition(temperature, pressure)

        def volume0() -> ArrayLike:
            return self.mrk_fluid.volume(temperature, pressure)

        def volume1() -> ArrayLike:
            return self.mrk_gas.volume(temperature, pressure)

        def volume2() -> ArrayLike:
            return self.mrk_fluid.volume(temperature, pressure)

        def volume3() -> ArrayLike:
            return self.mrk_liquid.volume(temperature, pressure)

        def volume4() -> ArrayLike:
            return self.mrk_fluid.volume(temperature, pressure)

        volume_funcs: list[Callable] = [volume0, volume1, volume2, volume3, volume4]

        volume: Array = lax.switch(condition, volume_funcs)
        # jax.debug.print("volume = {out}", out=volume)

        return volume


H2OMrkHolland91: RealGas = H2OMrkHP91()
"""H2O MRK that includes critical behaviour (also the liquid phase)"""
CO2_mrk_cs_holland91: RealGas = MRKCorrespondingStatesHP91.get_species("CO2_g")
"""CO2 MRK corresponding states :cite:p:`HP91`"""
CH4_mrk_cs_holland91: RealGas = MRKCorrespondingStatesHP91.get_species("CH4_g")
"""CH4 MRK corresponding states :cite:p:`HP91`"""
H2_mrk_cs_holland91: RealGas = MRKCorrespondingStatesHP91.get_species("H2_g_Holland")
"""H2 MRK corresponding states :cite:p:`HP91`"""
CO_mrk_cs_holland91: RealGas = MRKCorrespondingStatesHP91.get_species("CO_g")
"""CO MRK corresponding states :cite:p:`HP91`"""
N2_mrk_cs_holland91: RealGas = MRKCorrespondingStatesHP91.get_species("N2_g")
"""N2 MRK corresponding states :cite:p:`HP91`"""
S2_mrk_cs_holland11: RealGas = MRKCorrespondingStatesHP91.get_species("S2_g")
"""S2 MRK corresponding states :cite:p:`HP11`"""
H2S_mrk_cs_holland11: RealGas = MRKCorrespondingStatesHP91.get_species("H2S_g")
"""H2S MRK corresponding states :cite:p:`HP11`"""
H2O_mrk_fluid_holland91: RealGas = H2OMrkFluidHolland91
"""H2O MRK supercritical fluid :cite:p:`HP91`"""
H2O_mrk_gas_holland91: RealGas = H2OMrkGasHolland91
"""H2O MRK gas :cite:p:`HP91`"""
H2O_mrk_liquid_holland91: RealGas = H2OMrkLiquidHolland91
"""H2O MRK liquid :cite:p:`HP91`"""
H2O_mrk_gas_fluid_holland91: RealGas = H2OMrkGasFluid91()
"""H2O MRK for the gas and supercritical fluid :cite:p:`HP91`"""

coefficients_P: tuple[float, ...] = CorrespondingStatesUnitConverter.convert_virial_coefficients(
    (6.93054e-7, -8.38293e-8)
)
coefficients_sqrtP: tuple[float, ...] = (
    CorrespondingStatesUnitConverter.convert_virial_coefficients((-3.30558e-5, 2.30524e-6))
)
virial_compensation_corresponding_states: VirialCompensation = VirialCompensation(
    coefficients_P, coefficients_sqrtP, (0, 0), 0
)
"""Virial compensation for corresponding states :cite:p:`HP91{Table 2}`

In this case it appears `P0` is always zero, even though for the full CORK equations it determines
whether or not the virial contribution is added. The unit conversions to SI and pressure in bar 
mean that every virial coefficient has been multiplied by 1e-2 compared to the values in 
:cite:t:`HP91{Table 2}`.
"""

experimental_calibration_holland91: ExperimentalCalibration = ExperimentalCalibration(
    temperature_min=373,
    temperature_max=1873,
    pressure_min=1,
    pressure_max=50e3,
)
"""Experimental calibration for :cite:`HP91,HP11` models"""
experimental_calibration_holland98: ExperimentalCalibration = ExperimentalCalibration(
    temperature_min=373,
    temperature_max=1873,
    pressure_min=1,
    pressure_max=120e3,
)
"""Experimental calibration for :cite:`HP98` models"""

CH4_cork_cs_holland91: RealGas = CORK(
    CH4_mrk_cs_holland91, virial_compensation_corresponding_states, select_critical_data("CH4_g")
)
"""CH4 CORK corresponding states :cite:p:`HP91`"""
CH4_cork_cs_holland91_bounded: RealGas = CombinedRealGas.create(
    [CH4_cork_cs_holland91], [experimental_calibration_holland91]
)
"""CH4 CORK corresponding states bounded :cite:p:`HP91`"""
CO_cork_cs_holland91: RealGas = CORK(
    CO_mrk_cs_holland91, virial_compensation_corresponding_states, select_critical_data("CO_g")
)
"""CO CORK corresponding states :cite:p:`HP91`"""
CO_cork_cs_holland91_bounded: RealGas = CombinedRealGas.create(
    [CO_cork_cs_holland91], [experimental_calibration_holland91]
)
"""CO CORK corresponding states bounded :cite:p:`HP91`"""
CO2_cork_cs_holland91: RealGas = CORK(
    CO2_mrk_cs_holland91, virial_compensation_corresponding_states, select_critical_data("CO2_g")
)
"""CO2 CORK corresponding states :cite:p:`HP91`"""
CO2_cork_cs_holland91_bounded: RealGas = CombinedRealGas.create(
    [CO2_cork_cs_holland91], [experimental_calibration_holland91]
)
"""CO2 CORK corresponding states bounded :cite:p:`HP91`"""
H2_cork_cs_holland91: RealGas = CORK(
    H2_mrk_cs_holland91,
    virial_compensation_corresponding_states,
    select_critical_data("H2_g_Holland"),
)
"""H2 CORK corresponding states :cite:p:`HP91`"""
H2_cork_cs_holland91_bounded: RealGas = CombinedRealGas.create(
    [H2_cork_cs_holland91], [experimental_calibration_holland91]
)
"""H2 CORK corresponding states bounded :cite:p:`HP91`"""
H2S_cork_cs_holland11: RealGas = CORK(
    H2S_mrk_cs_holland11, virial_compensation_corresponding_states, select_critical_data("H2S_g")
)
"""H2S CORK corresponding states :cite:p:`HP91`"""
H2S_cork_cs_holland11_bounded: RealGas = CombinedRealGas.create(
    [H2S_cork_cs_holland11], [experimental_calibration_holland91]
)
"""H2S CORK corresponding states bounded :cite:p:`HP91`"""
N2_cork_cs_holland91: RealGas = CORK(
    N2_mrk_cs_holland91, virial_compensation_corresponding_states, select_critical_data("N2_g")
)
"""N2 CORK corresponding states :cite:p:`HP91`"""
N2_cork_cs_holland91_bounded: RealGas = CombinedRealGas.create(
    [N2_cork_cs_holland91], [experimental_calibration_holland91]
)
"""N2 CORK corresponding states bounded :cite:p:`HP91`"""
S2_cork_cs_holland11: RealGas = CORK(
    S2_mrk_cs_holland11, virial_compensation_corresponding_states, select_critical_data("S2_g")
)
"""S2 CORK corresponding states :cite:p:`HP91`"""
S2_cork_cs_holland11_bounded: RealGas = CombinedRealGas.create(
    [S2_cork_cs_holland11], [experimental_calibration_holland91]
)
"""S2 CORK corresponding states bounded :cite:p:`HP91`"""

dummy_critical_data: CriticalData = CriticalData(1.0, 1.0)
"""Dummy critical data

The full CO2 and H2O CORK models are not a corresponding states model, which can be reproduced by 
ignoring the scaling by the critical temperature and pressure, i.e. setting these quantities to 
unity.
"""
CO2_virial_compensation_holland91: VirialCompensation = VirialCompensation(
    FullUnitConverter.convert_virial_coefficients((1.33790e-2, -1.01740e-5), 1.0),
    FullUnitConverter.convert_virial_coefficients((-2.26924e-1, 7.73793e-5), 0.5),
    (0, 0),
    5000,
)
"""CO2 virial compensation :cite:p:`HP91`"""
CO2_cork_holland91: RealGas = CORK(
    CO2MrkHolland91, CO2_virial_compensation_holland91, dummy_critical_data
)
"""CO2 cork :cite:p:`HP91`"""
CO2_cork_holland91_bounded: RealGas = CombinedRealGas.create(
    [CO2_cork_holland91], [experimental_calibration_holland91]
)
"""CO2 cork bounded :cite:p:`HP91`"""

H2O_virial_compensation_holland91: VirialCompensation = VirialCompensation(
    FullUnitConverter.convert_virial_coefficients((-3.2297554e-3, 2.2215221e-6), 1.0),
    FullUnitConverter.convert_virial_coefficients((-3.025650e-2, -5.343144e-6), 0.5),
    (0, 0),
    2000,
)
"""H2O virial compensation :cite:p:`HP91`"""
H2O_cork_holland91: RealGas = CORK(
    H2OMrkHolland91, H2O_virial_compensation_holland91, dummy_critical_data
)
"""H2O cork :cite:p:`HP91`"""
H2O_cork_holland91_bounded: RealGas = CombinedRealGas.create(
    [H2O_cork_holland91], [experimental_calibration_holland91]
)
"""H2O cork bounded :cite:p:`HP91`"""
H2O_cork_gas_fluid_holland91: RealGas = CORK(
    H2O_mrk_gas_fluid_holland91, H2O_virial_compensation_holland91, dummy_critical_data
)
"""H2O cork for the gas and supercritical fluid :cite:p:`HP91`"""
H2O_cork_gas_fluid_holland91_bounded: RealGas = CombinedRealGas.create(
    [H2O_cork_gas_fluid_holland91], [experimental_calibration_holland91]
)
"""H2O cork for the gas and supercritical fluid bounded :cite:p:`HP91`"""

CO2_virial_compensation_holland98: VirialCompensation = VirialCompensation(
    FullUnitConverter.convert_virial_coefficients((5.40776e-3, -1.59046e-6), 1.0),
    FullUnitConverter.convert_virial_coefficients((-1.78198e-1, 2.45317e-5), 0.5),
    (0, 0),
    5000,
)
"""CO2 virial compensation :cite:p:`HP98`"""
CO2_cork_holland98: RealGas = CORK(
    CO2MrkHolland91, CO2_virial_compensation_holland98, dummy_critical_data
)
"""CO2 cork :cite:p:`HP98`"""
CO2_cork_holland98_bounded: RealGas = CombinedRealGas.create(
    [CO2_cork_holland98], [experimental_calibration_holland98]
)
"""CO2 cork bounded :cite:p:`HP98`"""

H2O_virial_compensation_holland98: VirialCompensation = VirialCompensation(
    FullUnitConverter.convert_virial_coefficients((1.9853e-3, 0), 1.0),
    FullUnitConverter.convert_virial_coefficients((-8.9090e-2, 0), 0.5),
    FullUnitConverter.convert_virial_coefficients((8.0331e-2, 0), 0.25),
    2000,
)
"""H2O virial compensation :cite:p:`HP98`"""
H2O_cork_holland98: RealGas = CORK(
    H2OMrkHolland91, H2O_virial_compensation_holland98, dummy_critical_data
)
"""H2O cork :cite:p:`HP98`"""
H2O_cork_holland98_bounded: RealGas = CombinedRealGas.create(
    [H2O_cork_holland98], [experimental_calibration_holland98]
)
"""H2O cork bounded :cite:p:`HP98`"""
H2O_cork_gas_fluid_holland98: RealGas = CORK(
    H2O_mrk_gas_fluid_holland91, H2O_virial_compensation_holland98, dummy_critical_data
)
"""H2O cork for the gas and supercritical fluid :cite:p:`HP98`"""
H2O_cork_gas_fluid_holland98_bounded: RealGas = CombinedRealGas.create(
    [H2O_cork_gas_fluid_holland98], [experimental_calibration_holland98]
)
"""H2O cork for the gas and supercritical fluid bounded :cite:p:`HP98`"""


def get_holland_eos_models() -> dict[str, RealGas]:
    """Gets a dictionary of EOS models

    The naming convention is as follows:
        [species]_[eos model]_[citation]

    'cs' refers to corresponding states.

    Returns:
        Dictionary of EOS models
    """
    eos_models: dict[str, RealGas] = {}
    eos_models["CH4_cork_cs_holland91"] = CH4_cork_cs_holland91_bounded
    eos_models["CO_cork_cs_holland91"] = CO_cork_cs_holland91_bounded
    eos_models["CO2_cork_holland91"] = CO2_cork_holland91_bounded
    eos_models["CO2_cork_holland98"] = CO2_cork_holland98_bounded
    eos_models["CO2_cork_cs_holland91"] = CO2_cork_cs_holland91_bounded
    eos_models["H2_cork_cs_holland91"] = H2_cork_cs_holland91_bounded
    eos_models["H2O_cork_holland91"] = H2O_cork_gas_fluid_holland91_bounded
    eos_models["H2O_cork_holland98"] = H2O_cork_gas_fluid_holland98_bounded
    eos_models["H2S_cork_cs_holland11"] = H2S_cork_cs_holland11_bounded
    eos_models["N2_cork_cs_holland91"] = N2_cork_cs_holland91_bounded
    eos_models["S2_cork_cs_holland11"] = S2_cork_cs_holland11_bounded

    return eos_models



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Redox buffers"""

from __future__ import annotations

from abc import abstractmethod
from typing import Type

import equinox as eqx
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from atmodeller.utilities import ExperimentalCalibration, unit_conversion

try:
    from typing import override  # type: ignore valid for Python 3.12+
except ImportError:
    from typing_extensions import override  # Python 3.11 and earlier


class RedoxBuffer(eqx.Module):
    """Redox buffer

    This must adhere to FugacityConstraintProtocol

    Args:
        log10_shift: Log10 shift relative to the buffer. Defaults to 0.
        evaluation_pressure: Pressure to evaluate the buffer at. Defaults to 1 bar. If None, then
            the total pressure will be used, but this can give rise to multiple solutions and
            should be used with caution.
    """

    log10_shift: ArrayLike = 0
    evaluation_pressure: ArrayLike | None = 1.0
    calibration: eqx.AbstractVar[ExperimentalCalibration] = eqx.field(init=False)

    @property
    def value(self) -> ArrayLike:
        return self.log10_shift

    @abstractmethod
    def convert_pressure_units(self, pressure: ArrayLike) -> ArrayLike:
        """Converts the pressure units

        Args:
            pressure: Pressure in bar

        Returns:
            Pressure in units appropriate for the calculation
        """

    @abstractmethod
    def log10_fugacity_buffer(self, temperature: ArrayLike, pressure: ArrayLike) -> ArrayLike:
        """Gets the log10 fugacity at the buffer

        Args:
            temperature: Temperature
            pressure: Pressure

        Returns:
            Log10 fugacity at the buffer
        """

    @eqx.filter_jit
    def get_scaled_pressure(self, pressure: ArrayLike) -> ArrayLike:
        """Gets the scaled pressure.

        Args:
            pressure: Pressure in bar

        Returns:
            Pressure in units appropriate for the calculation
        """
        if self.evaluation_pressure is not None:
            return self.convert_pressure_units(self.evaluation_pressure)
        else:
            return self.convert_pressure_units(pressure)

    @override
    @eqx.filter_jit
    def log10_fugacity(self, temperature: ArrayLike, pressure: ArrayLike) -> ArrayLike:
        """Gets the log10 fugacity

        Args:
            temperature: Temperature
            pressure: Pressure

        Returns:
            Log10 fugacity
        """
        return self.log10_fugacity_buffer(temperature, pressure) + self.log10_shift

    @override
    @eqx.filter_jit
    def log_fugacity(self, temperature: ArrayLike, pressure: ArrayLike) -> ArrayLike:
        """Gets the log fugacity

        Args:
            temperature: Temperature
            pressure: Pressure

        Returns:
            Log fugacity
        """
        return self.log10_fugacity(temperature, pressure) * jnp.log(10)


class IronWustiteBufferHirschmann08(RedoxBuffer):
    """Iron-wustite buffer :cite:p:`OP93,HGD08`

    Experimental calibration values are provided in the abstract of :cite:t:`HGD08`.

    Args:
        log10_shift: Log10 shift relative to the buffer. Defaults to zero.
        evaluation_pressure: Pressure to evaluate the buffer at. Defaults to 1 bar. If None, then
            the total pressure will be used, but this can give rise to multiple solutions and
            should be used with caution.
        calibration: Experimental calibration
    """

    calibration: ExperimentalCalibration = ExperimentalCalibration(
        pressure_max=27.5 * unit_conversion.GPa_to_bar
    )

    @override
    @eqx.filter_jit
    def convert_pressure_units(self, pressure: ArrayLike) -> ArrayLike:
        """Units are bar"""
        return pressure

    @override
    @eqx.filter_jit
    def log10_fugacity_buffer(self, temperature: ArrayLike, pressure: ArrayLike) -> ArrayLike:
        """Gets the log10 fugacity

        Args:
            temperature: Temperature
            pressure: Pressure

        Returns:
            Log10 fugacity
        """
        scaled_pressure: ArrayLike = self.get_scaled_pressure(pressure)
        log10_fugacity_buffer: Array = (
            -0.8853 * jnp.log(temperature)
            - 28776.8 / temperature
            + 14.057
            + 0.055 * (scaled_pressure - 1) / temperature
        )

        return log10_fugacity_buffer


class IronWustiteBufferHirschmann21(RedoxBuffer):
    """Iron-wustite buffer :cite:p:`H21`

    Regarding the calibration, :cite:t:`H21` states that: 'It extrapolates smoothly to higher
    temperature, though not calibrated above 3000 K. Extrapolation to lower temperatures (<1000 K)
    or higher pressures (>100 GPa) is not recommended.'

    Args:
        log10_shift: Log10 shift relative to the buffer. Defaults to zero.
        evaluation_pressure: Pressure to evaluate the buffer at. Defaults to 1 bar. If None, then
            the total pressure will be used, but this can give rise to multiple solutions and
            should be used with caution.
        calibration: Experimental calibration
        a-h: Coefficients
    """

    calibration: ExperimentalCalibration = ExperimentalCalibration(
        temperature_min=1000, pressure_max=100 * unit_conversion.GPa_to_bar
    )
    a: tuple[float, ...] = (6.844864, 1.175691e-1, 1.143873e-3, 0, 0)
    b: tuple[float, ...] = (5.791364e-4, -2.891434e-4, -2.737171e-7, 0.0, 0.0)
    c: tuple[float, ...] = (-7.971469e-5, 3.198005e-5, 0, 1.059554e-10, 2.014461e-7)
    d: tuple[float, ...] = (-2.769002e4, 5.285977e2, -2.919275, 0, 0)
    e: tuple[float, ...] = (8.463095, -3.000307e-3, 7.213445e-5, 0, 0)
    f: tuple[float, ...] = (1.148738e-3, -9.352312e-5, 5.161592e-7, 0, 0)
    g: tuple[float, ...] = (-7.448624e-4, -6.329325e-6, 0, -1.407339e-10, 1.830014e-4)
    h: tuple[float, ...] = (-2.782082e4, 5.285977e2, -8.473231e-1, 0, 0)

    @override
    @eqx.filter_jit
    def convert_pressure_units(self, pressure: ArrayLike) -> ArrayLike:
        """Units are GPa"""
        return pressure * unit_conversion.bar_to_GPa

    @eqx.filter_jit
    def _evaluate_m(self, pressure: ArrayLike, coefficients: tuple[float, ...]) -> Array:
        """Evaluates an m parameter

        Args:
            pressure: Pressure in GPa
            coefficients: Coefficients

        Return:
            m parameter
        """
        m: Array = (
            coefficients[4] * jnp.power(pressure, 0.5)
            + coefficients[0]
            + coefficients[1] * pressure
            + coefficients[2] * jnp.power(pressure, 2)
            + coefficients[3] * jnp.power(pressure, 3)
        )

        return m

    @eqx.filter_jit
    def _evaluate_fO2(
        self,
        temperature: ArrayLike,
        pressure: ArrayLike,
        coefficients: tuple[tuple[float, ...], ...],
    ) -> Array:
        """Evaluates the fO2

        Args:
            temperature: Temperature in K
            pressure: Pressure in GPa
            coefficients: Coefficients

        Returns:
            log10fO2
        """
        log10fO2: Array = (
            self._evaluate_m(pressure, coefficients[0])
            + self._evaluate_m(pressure, coefficients[1]) * temperature
            + self._evaluate_m(pressure, coefficients[2]) * temperature * jnp.log(temperature)
            + self._evaluate_m(pressure, coefficients[3]) / temperature
        )

        return log10fO2

    @eqx.filter_jit
    def _fcc_bcc_iron(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """log10fO2 for fcc and bcc iron

        Args:
            temperature: Temperature in K
            pressure: Pressure in GPa

        Return:
            log10fO2 for fcc and bcc iron
        """
        log10fO2: Array = self._evaluate_fO2(
            temperature, pressure, (self.a, self.b, self.c, self.d)
        )

        return log10fO2

    @eqx.filter_jit
    def _hcp_iron(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """log10fO2 for hcp iron

        Args:
            temperature: Temperature in K
            pressure: Pressure in GPa

        Return:
            log10fO2 for hcp iron
        """
        log10fO2: Array = self._evaluate_fO2(
            temperature, pressure, (self.e, self.f, self.g, self.h)
        )

        return log10fO2

    @eqx.filter_jit
    def _use_hcp(self, temperature: ArrayLike, pressure: ArrayLike) -> Array:
        """Check to use hcp iron formulation for fO2

        Args:
            temperature: Temperature in K
            pressure: Pressure in GPa

        Returns:
            True/False whether to use the hcp iron formulation
        """
        x: tuple[float, ...] = (-18.64, 0.04359, -5.069e-6)
        threshold: Array = x[2] * jnp.power(temperature, 2) + x[1] * temperature + x[0]

        return jnp.array(pressure) > threshold

    @override
    @eqx.filter_jit
    def log10_fugacity_buffer(self, temperature: ArrayLike, pressure: ArrayLike) -> ArrayLike:
        """Gets the log10 fugacity

        Args:
            temperature: Temperature
            pressure: Pressure

        Returns:
            Log10 fugacity
        """
        scaled_pressure: ArrayLike = self.get_scaled_pressure(pressure)

        def hcp_case() -> Array:
            return self._hcp_iron(temperature, scaled_pressure)

        def fcc_bcc_case() -> Array:
            return self._fcc_bcc_iron(temperature, scaled_pressure)

        buffer_value: Array = jnp.where(
            self._use_hcp(temperature, scaled_pressure), hcp_case(), fcc_bcc_case()
        )

        return buffer_value


class IronWustiteBufferHirschmann(RedoxBuffer):
    """Composite iron-wustite buffer using :cite:t:`OP93,HGD08` and :cite:t:`H21`

    Args:
        log10_shift: Log10 shift relative to the buffer. Defaults to 0.
        evaluation_pressure: Pressure to evaluate the buffer at. Defaults to 1 bar. If None, then
            the total pressure will be used, but this can give rise to multiple solutions and
            should be used with caution.
        calibration: Experimental calibration
    """

    calibration: ExperimentalCalibration = ExperimentalCalibration(
        pressure_max=100 * unit_conversion.GPa_to_bar
    )
    low_temperature_buffer: RedoxBuffer = eqx.field(init=False)
    high_temperature_buffer: RedoxBuffer = eqx.field(init=False)

    def __post_init__(self):
        self.low_temperature_buffer = IronWustiteBufferHirschmann08(
            self.log10_shift, self.evaluation_pressure
        )
        self.high_temperature_buffer = IronWustiteBufferHirschmann21(
            self.log10_shift, self.evaluation_pressure
        )

    # Not used for a composite redox buffer but required by the interface
    @override
    @eqx.filter_jit
    def convert_pressure_units(self, pressure: ArrayLike) -> ArrayLike:
        """Units are bar"""
        return pressure

    @eqx.filter_jit
    def _use_low_temperature(self, temperature: ArrayLike) -> Array:
        """Check to use the low temperature buffer for fO2

        Args:
            temperature: Temperature

        Returns:
            True/False whether to use the low temperature formulation
        """
        return jnp.asarray(temperature) < self.high_temperature_buffer.calibration.temperature_min

    @override
    @eqx.filter_jit
    def log10_fugacity_buffer(self, temperature: ArrayLike, pressure: ArrayLike) -> ArrayLike:
        """Gets the log10 fugacity at the buffer

        Args:
            temperature: Temperature
            pressure: Pressure

        Returns:
            Log10 fugacity at the buffer
        """

        def low_temperature_case() -> ArrayLike:
            return self.low_temperature_buffer.log10_fugacity_buffer(temperature, pressure)

        def high_temperature_case() -> ArrayLike:
            return self.high_temperature_buffer.log10_fugacity_buffer(temperature, pressure)

        buffer_value: Array = jnp.where(
            self._use_low_temperature(temperature),
            low_temperature_case(),
            high_temperature_case(),
        )

        return buffer_value


IronWustiteBuffer: Type[RedoxBuffer] = IronWustiteBufferHirschmann
